!!
!! dangling.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!!!!! This file contains F77 functions for reporting the
!!!!! 3' and 5' dangling end energies, based on the input
!!!!! bases
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! dangle3() returns the 3' dangling end energy according to
! the following configuration diagram, where 1,2,3 are the
! input bases base1,base2,base3 respectively. These bases are
! preset to base id by the caller of this subroutine according to the
! following scheme: (A, C, G, U)=(1,2,3,4).
!

!       data input for ed3(1,2,3) for 3' dangle
!       5' --> 3'
!          23
!          1
!       3' <-- 5'

      function dangle3(base1, base2, base3, pardir, pardirlen)

        double precision dangle3
        integer pardirlen
        integer base1, base2, base3
        character*(*) pardir

        integer buflen, maxfnlen
        parameter(buflen=1000)
        parameter(maxfnlen=200)

        external data

        double precision es(4, 4, 4, 4),
     &       ed5(4,4,4), ed5final(4,4,4), ed3(4,4,4),
     &       ehs(4, 4, 4, 4), eis(4, 4, 4, 4),
     &       e21s(4,4,4,4,4,4,4),  e21d(4,4,4,4,4,4,4),
     &       ei22a(4,4,4,4,4,4,4), ei22c(4,4,4,4,4,4,4),
     &       ei22g(4,4,4,4,4,4,4), ei22u(4,4,4,4,4,4,4),
     &       ei11(4,4,4,4,4,4),
     &       eh3(4,4,4,4,4), eh4(4,4,4,4,4,4),
     &       eh(30), eb(30), ei(30)

        double precision  eh3ex(4,4,4,4,4), eh4ex(4,4,4,4,4,4),
     &        esex(4, 4, 4, 4), ed3ex(4,4,4), ed5ex(4,4,4),
     &       ehsex(4, 4, 4, 4), eisex(4, 4, 4, 4),
     &       ehex(30), ebex(30), eiex(30),
     &        ei11ex(4,4,4,4,4,4),
     &       ei22aex(4,4,4,4,4,4,4), ei22cex(4,4,4,4,4,4,4),
     &       ei22gex(4,4,4,4,4,4,4), ei22uex(4,4,4,4,4,4,4),
     &       e21sex(4,4,4,4,4,4,4),  e21dex(4,4,4,4,4,4,4),
     &       aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        common /stack/ es /dangle/ ed3, ed5, ed5final /HBI/ eh, eb, ei
     &       /Hstack/ ehs /Istack/ eis /int11/ ei11
     &       /int21/ e21s, e21d/ int22
     &       / ei22a,  ei22c, ei22g, ei22u
     &       /tetra/ eh4 / tri /eh3

        common /stackex/ esex /dangleex/ ed3ex, ed5ex
     &       /HBIex/ ehex, ebex, eiex
     &       /Hstackex/ ehsex /Istackex/ eisex /int11ex/ ei11ex
     &       /int21ex/ e21sex, e21dex/ int22ex
     &       / ei22aex,  ei22cex, ei22gex, ei22uex
     &       /tetraex/ eh4ex / triex /eh3ex
     &       /other/aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        dangle3=0.0d0
        !    2006/11/16 (Clarence): error checking on buflen to make sure
        !  pardirlen+maxfnlen would not
        !  be greater than the buflen defined above.
        !
        if (pardirlen+maxfnlen .gt. buflen) then
          write(0, *)  " Error: length of parameter directory too long"
          go to 900
        end if

        !  make sure the input bases have valid base id (1-4)
        if (base1 .lt. 1 .or. base1 .gt. 4) then
          write(0, *)  " Error: invalid base ID for base 1"
          go to 900
        end if
        if (base2 .lt. 1 .or. base2 .gt. 4) then
          write(0, *)  " Error: invalid base ID for base 2"
          go to 900
        end if
        if (base3 .lt. 1 .or. base3 .gt. 4) then
          write(0, *)  " Error: invalid base ID for base 3"
          go to 900
        end if

        call data(pardir, pardirlen)

        dangle3 = ed3(base1, base2, base3)

 900    return
      end

      !
      ! dangle5() returns the 5' dangling end energy according to
      ! the following configuration diagram, where 1,2,3 are the
      ! input bases base1,base2,base3 respectively. These bases are
      ! preset to base id by the caller of this subroutine according to the
      ! following scheme: (A, C, G, U)=(1,2,3,4).
      !

      !       data input for ed5(1,2,3) for 5' dangle
      !       5' --> 3'
      !          2
      !          13
      !       3' <-- 5'

      function dangle5(base1, base2, base3, pardir, pardirlen)

        double precision dangle5
        integer pardirlen
        integer base1, base2, base3
        character*(*) pardir

        integer buflen, maxfnlen
        parameter(buflen=1000)
        parameter(maxfnlen=200)

        external data
        double precision es(4, 4, 4, 4),
     &       ed5(4,4,4), ed5final(4,4,4), ed3(4,4,4),
     &       ehs(4, 4, 4, 4), eis(4, 4, 4, 4),
     &       e21s(4,4,4,4,4,4,4),  e21d(4,4,4,4,4,4,4),
     &       ei22a(4,4,4,4,4,4,4), ei22c(4,4,4,4,4,4,4),
     &       ei22g(4,4,4,4,4,4,4), ei22u(4,4,4,4,4,4,4),
     &       ei11(4,4,4,4,4,4),
     &       eh3(4,4,4,4,4), eh4(4,4,4,4,4,4),
     &       eh(30), eb(30), ei(30)

        double precision  eh3ex(4,4,4,4,4), eh4ex(4,4,4,4,4,4),
     &        esex(4, 4, 4, 4), ed3ex(4,4,4), ed5ex(4,4,4),
     &       ehsex(4, 4, 4, 4), eisex(4, 4, 4, 4),
     &       ehex(30), ebex(30), eiex(30),
     &        ei11ex(4,4,4,4,4,4),
     &       ei22aex(4,4,4,4,4,4,4), ei22cex(4,4,4,4,4,4,4),
     &       ei22gex(4,4,4,4,4,4,4), ei22uex(4,4,4,4,4,4,4),
     &       e21sex(4,4,4,4,4,4,4),  e21dex(4,4,4,4,4,4,4),
     &       aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        common /stack/ es /dangle/ ed3, ed5, ed5final /HBI/ eh, eb, ei
     &       /Hstack/ ehs /Istack/ eis /int11/ ei11
     &       /int21/ e21s, e21d/ int22
     &       / ei22a,  ei22c, ei22g, ei22u
     &       /tetra/ eh4 / tri /eh3

        common /stackex/ esex /dangleex/ ed3ex, ed5ex
     &       /HBIex/ ehex, ebex, eiex
     &       /Hstackex/ ehsex /Istackex/ eisex /int11ex/ ei11ex
     &       /int21ex/ e21sex, e21dex/ int22ex
     &       / ei22aex,  ei22cex, ei22gex, ei22uex
     &       /tetraex/ eh4ex / triex /eh3ex
     &       /other/aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        dangle5=0.0d0
        !    2006/11/16 (Clarence): error checking on buflen to make sure
        !  pardirlen+maxfnlen would not
        !  be greater than the buflen defined above.
        !
        if (pardirlen+maxfnlen .gt. buflen) then
          write(0, *)  " Error: length of parameter directory too long"
          go to 900
        end if

        !  make sure the input bases have valid base id (1-4)
        if (base1 .lt. 1 .or. base1 .gt. 4) then
          write(0, *)  " Error: invalid base ID for base 1"
          go to 900
        end if
        if (base2 .lt. 1 .or. base2 .gt. 4) then
          write(0, *)  " Error: invalid base ID for base 2"
          go to 900
        end if
        if (base3 .lt. 1 .or. base3 .gt. 4) then
          write(0, *)  " Error: invalid base ID for base 3"
          go to 900
        end if

        call data(pardir, pardirlen)

        dangle5 = ed5(base1, base2, base3)

 900    return
      end

      include 'data.f'
