!!
!! diffstab.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!!!!! This is the F77 function for computing the DSSE of a siRNA
!!!!! duplex. Unlike the original code
!!!!! in Sfold for calculating this total, this program does not limit
!!!!! the base-pairing to W-C pairs or the length of the siRNAs to 21
!!!!! (19+2-nt overhang).
!!!!! This code is modified from duplexstab.f
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! sense and antisense contain the two strands of the duplex. they are
! preset to base id by the caller of this subroutine according to the
! following scheme: (A, C, G, U)=(1,2,3,4). nsirna is the length of
! each strand
!
      function diffstab(sense, antisense, nsirna, pardir, pardirlen)

        double precision diffstab
        integer nsirna, pardirlen
        integer sense(nsirna), antisense(nsirna)
        character*(*) pardir

!  DSSE: differential stability of siRNA duplex ends
!  dg5s: stability of 5' end of sense strand
!  dg5as: stability of 5' end of antisense strand
!  tphb: terminal A-U penalty
        double precision DSSE, dg5s, dg5as, tphb

!  counter variables
        integer i

        integer buflen, maxfnlen
        parameter(buflen=1000)
        parameter(maxfnlen=200)

        external data

        double precision es(4, 4, 4, 4),
     &       ed5(4,4,4), ed5final(4,4,4), ed3(4,4,4),
     &       ehs(4, 4, 4, 4), eis(4, 4, 4, 4),
     &       e21s(4,4,4,4,4,4,4),  e21d(4,4,4,4,4,4,4),
     &       ei22a(4,4,4,4,4,4,4), ei22c(4,4,4,4,4,4,4),
     &       ei22g(4,4,4,4,4,4,4), ei22u(4,4,4,4,4,4,4),
     &       ei11(4,4,4,4,4,4),
     &       eh3(4,4,4,4,4), eh4(4,4,4,4,4,4),
     &       eh(30), eb(30), ei(30)

        double precision  eh3ex(4,4,4,4,4), eh4ex(4,4,4,4,4,4),
     &        esex(4, 4, 4, 4), ed3ex(4,4,4), ed5ex(4,4,4),
     &       ehsex(4, 4, 4, 4), eisex(4, 4, 4, 4),
     &       ehex(30), ebex(30), eiex(30),
     &        ei11ex(4,4,4,4,4,4),
     &       ei22aex(4,4,4,4,4,4,4), ei22cex(4,4,4,4,4,4,4),
     &       ei22gex(4,4,4,4,4,4,4), ei22uex(4,4,4,4,4,4,4),
     &       e21sex(4,4,4,4,4,4,4),  e21dex(4,4,4,4,4,4,4),
     &       aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        common /stack/ es /dangle/ ed3, ed5, ed5final /HBI/ eh, eb, ei
     &       /Hstack/ ehs /Istack/ eis /int11/ ei11
     &       /int21/ e21s, e21d/ int22
     &       / ei22a,  ei22c, ei22g, ei22u
     &       /tetra/ eh4 / tri /eh3

        common /stackex/ esex /dangleex/ ed3ex, ed5ex
     &       /HBIex/ ehex, ebex, eiex
     &       /Hstackex/ ehsex /Istackex/ eisex /int11ex/ ei11ex
     &       /int21ex/ e21sex, e21dex/ int22ex
     &       / ei22aex,  ei22cex, ei22gex, ei22uex
     &       /tetraex/ eh4ex / triex /eh3ex
     &       /other/aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        data tphb / 0.5d0/

        diffstab = 0.0d0
!    2006/11/16 (Clarence): error checking on buflen to make sure
!  pardirlen+maxfnlen would not
!  be greater than the buflen defined above.
        if (pardirlen+maxfnlen .gt. buflen) then
          write(0, *)  " Error: length of parameter directory too long"
          go to 900
        end if

!  make sure the input duplex has valid base id (1-4)
        do i=1, nsirna
          if (sense(i) .lt. 1 .or. sense(i) .gt. 4) then
            write(0, *)  " Error: invalid base in sense strand"
            go to 900
          end if

          if (antisense(i) .lt. 1 .or. antisense(i) .gt. 4) then
            write(0, *)  " Error: invalid base in antisense strand"
            go to 900
          end if
        end do

        call data(pardir, pardirlen)

!       compute stability of 5' end of both sense and antisense strand
!       by summing of 4 stacks and 3' dangling dT
!       siRNA base position=i+2 (for dinucleotide leader)

        dg5s = 0.d0
!  4 stacking pairs (5 bases) at the 5' end of the sense strand
        do i=1, 4
          dg5s = dg5s + es(sense(i), sense(i+1), antisense(nsirna-i-2),
     &         antisense(nsirna-i-1))
        end do
!  3' dangling end at the 5' end of the sense strand
        dg5s = dg5s + ed3(sense(1), antisense(nsirna-2), antisense(nsirna-1))

!               add terminal A-U penalty (not used in Zamore calculation!)
        if ((sense(1) .eq. 1 .and. antisense(nsirna-2) .eq. 4) .or.
     &         (sense(1) .eq. 4 .and. antisense(nsirna-2) .eq. 1)) then
        dg5s = dg5s + tphb
        end if

        dg5as = 0.d0
!  4 stacking pairs (5 bases) at the 5' end of the antisense strand
        do i=1, 4
          dg5as = dg5as + es(antisense(i), antisense(i+1), sense(nsirna-i-2),
     &         sense(nsirna-i-1))
        end do
!  3' dangling end at the 5' end of the antisense strand
        dg5as = dg5as + ed3(antisense(1), sense(nsirna-2), sense(nsirna-1))

!               add terminal A-U penalty (not used in Zamore calculation!)
        if ((antisense(1) .eq. 1 .and. sense(nsirna-2) .eq. 4) .or.
     &         (antisense(1) .eq. 4 .and. sense(nsirna-2) .eq. 1)) then
        dg5as = dg5as + tphb
        end if

        DSSE = dg5as - dg5s

        diffstab = DSSE
 900    return
      end

      include 'data.f'
