!!
!! discrete.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!    2006/10/17 (Clarence): updated the Mersenne Twister to its latest
!       version (2002/1/26) with improved initialization. This version
!  does not contain the GPL terms. See copyright notice at the
!  end of this file. Fortran 77 code for this MT implementation
!  is in file mt19937ar.f
!

!  This is the function for generating  index  for a discrete
!  distribution (finite, trucate otherwise), disc(nmass, p)
!      nmass: # of points, p: probability vector
!  using Mersenne Twister
!  random number generator for unif(0,1) (2 subroutines
!  init_genrand() and genrand_real1() from MT are called)
!  see copyright notice at the end of this file
!  potential problem for distribution: using non standard intrinsics,
!  ishft(i,n),iand (i,j), ior  (i,j), ieor (i,j)
!  which are available on Sun Fortran, and in Compaq Fortran
!  Use the inversion (search) method for converting U(0,1) to any discrete
!  see p 71 of Ripley's book, although Alias method (Walker 1974, 77)
!  and its variant (Kronmal et al 1979) are faster per call (p 76 Ripley)
!  The setup time is longer (O(N^(2) or O(N)), for sampling, we only
!  need one call for sampling one base pair, and the search is linear
!  itime and getpid need to be replaced by time abd process id call
!  on the system, and getpid is not needed for running on one processor
!  it is only useful for preventing repeat on parellel run on multi-CPUs
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine discrete(nmass, p, index)
        integer nmass, j, index
        double precision unif, p(nmass), cdf

!  it's important to specify the function return type as follows,
!  or else the numbers will be computed wrongly (esp under Solaris
!  Sparc)
        double precision genrand_real1

        unif=genrand_real1()

        cdf=0.d0
        do j=1, nmass
          cdf=cdf+p(j)
          if (unif .le. cdf) then
            index=j
            go to 10
          end if
        end do

10      continue

        return
      end

      include 'mt19937ar.f'

!
!  A C-program for MT19937, with initialization improved 2002/1/26.
!  Coded by Takuji Nishimura and Makoto Matsumoto.
!
!  Before using, initialize the state by using init_genrand(seed)
!  or init_by_array(init_key, key_length).
!
!  Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
!  All rights reserved.
!  Copyright (C) 2005, Mutsuo Saito,
!  All rights reserved.
!
!  Redistribution and use in source and binary forms, with or without
!  modification, are only permitted provided that the following conditions
!  are met:
!
!    1. Redistributions of source code must retain the above copyright
!       notice, this list of conditions and the following disclaimer.
!
!    2. Redistributions in binary form must reproduce the above copyright
!       notice, this list of conditions and the following disclaimer in the
!       documentation and/or other materials provided with the distribution.
!
!    3. The names of its contributors may not be used to endorse or promote
!       products derived from this software without specific prior written
!       permission.
!
!  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
!  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
!  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
!  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
!  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
!  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
!  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
!  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
!  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
!  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
!  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
!
!  Any feedback is very welcome.
!  http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
!  email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
!
!-----------------------------------------------------------------------
!  FORTRAN77 translation by Tsuyoshi TADA. (2005/12/19)
!
!     ---------- initialize routines ----------
!  subroutine init_genrand(seed): initialize with a seed
!  subroutine init_by_array(init_key,key_length): initialize by an array
!
!     ---------- generate functions ----------
!  integer function genrand_int32(): signed 32-bit integer
!  integer function genrand_int31(): unsigned 31-bit integer
!  double precision function genrand_real1(): [0,1] with 32-bit resolution
!  double precision function genrand_real2(): [0,1) with 32-bit resolution
!  double precision function genrand_real3(): (0,1) with 32-bit resolution
!  double precision function genrand_res53(): (0,1) with 53-bit resolution
!
!  This program uses the following non-standard intrinsics.
!    ishft(i,n): If n>0, shifts bits in i by n positions to left.
!                If n<0, shifts bits in i by n positions to right.
!    iand (i,j): Performs logical AND on corresponding bits of i and j.
!    ior  (i,j): Performs inclusive OR on corresponding bits of i and j.
!    ieor (i,j): Performs exclusive OR on corresponding bits of i and j.
!
