!!
!! intstabprof.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!!!!! This is the F77 subroutine for computing the internal stability
!!!!! profile of siRNA duplex
!!!!! as requested by Anton McCaffrey. Unlike the original code
!!!!! in Sfold for calculating this total, this program does not limit
!!!!! the base-pairing to W-C pairs or the length of the siRNAs to 21
!!!!! (19+2-nt overhang). Internal stability for each of the 5'
!!!!! antisense strand positions (from 1 up to nsirna-2) is stored in
!!!!! the allocated array intdg.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! sense and antisense contain the two strands of the duplex. they are
! preset to base id by the caller of this subroutine according to the
! following scheme: (A, C, G, U)=(1,2,3,4). nsirna is the length of
! each strand. intdg is the allocated array of memory for storing the
! internal stability at each position of the duplex (i.e. length of
! array = nsirna - 2)
!

      subroutine intstabprof(sense, antisense, nsirna, pardir, pardirlen,
     &       intdg)

        integer nsirna, pardirlen
        integer sense(nsirna), antisense(nsirna)
        double precision intdg(nsirna-2)
        character*(*) pardir

!  intdg: array containing stabilities of all pentamer subsequences
!  tphb: terminal A-U penalty
!
        double precision tphb

!  counter variables
        integer i, j

        integer buflen, maxfnlen
        parameter(buflen=1000)
        parameter(maxfnlen=200)

        external data

        double precision es(4, 4, 4, 4), 
     &       ed5(4,4,4), ed5final(4,4,4), ed3(4,4,4),
     &       ehs(4, 4, 4, 4), eis(4, 4, 4, 4),
     &       e21s(4,4,4,4,4,4,4),  e21d(4,4,4,4,4,4,4),
     &       ei22a(4,4,4,4,4,4,4), ei22c(4,4,4,4,4,4,4),
     &       ei22g(4,4,4,4,4,4,4), ei22u(4,4,4,4,4,4,4),
     &       ei11(4,4,4,4,4,4),
     &       eh3(4,4,4,4,4), eh4(4,4,4,4,4,4),
     &       eh(30), eb(30), ei(30)

        double precision  eh3ex(4,4,4,4,4), eh4ex(4,4,4,4,4,4),
     &        esex(4, 4, 4, 4), ed3ex(4,4,4), ed5ex(4,4,4),
     &       ehsex(4, 4, 4, 4), eisex(4, 4, 4, 4),
     &       ehex(30), ebex(30), eiex(30),
     &        ei11ex(4,4,4,4,4,4),
     &       ei22aex(4,4,4,4,4,4,4), ei22cex(4,4,4,4,4,4,4),
     &       ei22gex(4,4,4,4,4,4,4), ei22uex(4,4,4,4,4,4,4),
     &       e21sex(4,4,4,4,4,4,4),  e21dex(4,4,4,4,4,4,4),
     &       aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        common /stack/ es /dangle/ ed3, ed5, ed5final /HBI/ eh, eb, ei
     &       /Hstack/ ehs /Istack/ eis /int11/ ei11
     &       /int21/ e21s, e21d/ int22
     &       / ei22a,  ei22c, ei22g, ei22u
     &       /tetra/ eh4 / tri /eh3

        common /stackex/ esex /dangleex/ ed3ex, ed5ex
     &       /HBIex/ ehex, ebex, eiex
     &       /Hstackex/ ehsex /Istackex/ eisex /int11ex/ ei11ex
     &       /int21ex/ e21sex, e21dex/ int22ex
     &       / ei22aex,  ei22cex, ei22gex, ei22uex
     &       /tetraex/ eh4ex / triex /eh3ex
     &       /other/aex, bex, cex, slopeex,
     &       tphbex, hguex, h3cex, haex, hbex, iasyex, ex3

        data tphb / 0.5d0/


!    2006/11/16 (Clarence): error checking on buflen to make sure
!  pardirlen+maxfnlen would not
!  be greater than the buflen defined above.
!
        if (pardirlen+maxfnlen .gt. buflen) then
          write(0, *)  " Error: length of parameter directory too long"
          go to 900
        end if


!  make sure the input duplex has valid base id (1-4)
        do i=1, nsirna
          if (sense(i) .lt. 1 .or. sense(i) .gt. 4) then
            write(0, *)  " Error: invalid base in sense strand"
            go to 900
          end if

          if (antisense(i) .lt. 1 .or. antisense(i) .gt. 4) then
            write(0, *)  " Error: invalid base in antisense strand"
            go to 900
          end if
        end do

        call data(pardir, pardirlen)


!  we calculate internal stability from 5' to 3' on the antisense
!  strand of the given duplex.
!  calculate the sum of 4 consecutive BP stacks starting at
!  position i, for i ranging from 1 to nsirna-6. For nsirna-5 up to
!  nsirna-2, we do not have enough BP stacks on the 5'->3' direction.
!  Thus, we reverse the direction of shifting so that
!  intdg(i)=intdg(i-4) for i from nsirna-5 to nsirna-2.
!  Dangling and A-U terminal penalties are included for intdg(1) and
!  intdg(nsirna-2).
!
        do i=1, nsirna-6
          intdg(i) = 0.d0
          do j=0, 3
            intdg(i) = intdg(i) + es(antisense(i+j), antisense(i+j+1),
     &         sense(nsirna-i-j-2), sense(nsirna-i-j-1))
          end do
        end do

!  reverse shifting for the last 4 positions
        do i=nsirna-5, nsirna-2
          intdg(i) = intdg(i-4)
        end do


!  3' dangling end at the 5' end of the antisense strand
        intdg(1) = intdg(1) + ed3(antisense(1), sense(nsirna-2),
     &       sense(nsirna-1))

!               add terminal A-U penalty (not used in Zamore calculation!)
        if ((antisense(1) .eq. 1 .and. sense(nsirna-2) .eq. 4) .or.
     &         (antisense(1) .eq. 4 .and. sense(nsirna-2) .eq. 1)) then
        intdg(1) = intdg(1) + tphb
        end if



!  3' dangling end at the 5' end of the sense strand
        intdg(nsirna-2) = intdg(nsirna-2) + ed3(sense(1), antisense(nsirna-2),
     &       antisense(nsirna-1))

!               add terminal A-U penalty (not used in Zamore calculation!)
        if ((sense(1) .eq. 1 .and. antisense(nsirna-2) .eq. 4) .or.
     &         (sense(1) .eq. 4 .and. antisense(nsirna-2) .eq. 1)) then
        intdg(nsirna-2) = intdg(nsirna-2) + tphb
        end if

 900    return
      end

      include 'data.f'
