!    compaq version, use 4 7D arrays ei22a,ei22ac,ei22ag, ei22aU
!!!!    Rescale version to avoid overflow, u0**n is est or exact u(1,n)
!!        pscale(i)=1/(u0**i), =0 for large i (underflow)
!!!!!!! This is the subroutine to compute the restricted  partition functions
!    (PF) u(i,j), up(i,j) (for base pair i.j) for general loop energies
!    for now: (a) size of I loop <= 30 (to maintain o(n^{3}) )
!         (b) single base stacking and updated Turner data used
!         (c) log-linear for H,B-loop > 30, with slope=1.079
!         (d) coaxial stacking unaddressed
!    Scale=(kcal/mol)/RT=1.6225072, R-gas constant, T=c+273.15, here c=37
!    linearity is assumed for multi-loop: em=a+b*ns+c*nh,
!    ns=# of single bases, nh=# of helices
!
!    uph, ups, upbr, upbl, upi, upm are components of up for
!    upbr: single basase at 3' (right side of loop), upbl: 5', left side
!    h-loop, stacking, b-loop, I-loop, and M-loop
!    u1(n, n) is a special PF related to m-loop penalties,needed for sampling
!    s1(n,n), s2(n,n), s3(n,n) are auxiliary arrays for reducing o(n^(4))
!        to o(n^{3}) by storing  more n^{2} arrays
!    note: only n(n+1)/2+ arrays needed, for i<=j, and i=j+1 for u
!    and i=j+1, j+2 for u1 (check all variable initials)
!    inc: increment j-i
!    tphb: terminal penalty for A-u,U-A, G-A, A-G for hydrogen bonds
!    h3c: penalty for 3 c in  triloop; ha, hb, coef. for >= 4C in h-loop
!    iasy: asym penalty;
!    see data.f for format of es, ehs, eis, ed3, ed5, ed5final, e21s, e21d, ei22
!    now use all energt in exponential form, i.e, exp(-E/RT)!!!
!
!    2004/09/21 (Clarence): updated to eliminate the cases for isolated
!    non-canonical bps. this is done by assigning -999.d0 to
!    all possible non-canonical bp entries in up(i,j) for j>=i+4, and
!    then by adding an if case inside partf to skip the calculation
!    of up(i,j) and set it directly to 0.d0. see individual statements
!    for details
!
!    2004/12/20 (Clarence): updated parameter list to include local folding
!    length lflen. this is used to change the big loop indices to reduce
!    the actual iteration it's run and thus reduce runtime
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine partf(n, numseq, pscale,  u, up,  u1, s1, s2, s3, lflen)
        integer n, numseq(n), h, l, inc, i, j , lflen
        double precision   pscale(n),
     &    u(n+1, n+1), up(n, n), u1(n+1, n+1), s1(n,n), s2(n,n), s3(n,n),
     &    uph, ups, upbr, upbl, upi, upm, sum1, sum2, sum3, stack3, termp,
     &    c3p, prod1, prod2, scaleterm, hpeex, tmp
        logical augu
!
        double precision  eh3ex(4,4,4,4,4), eh4ex(4,4,4,4,4,4),
     &         esex(4, 4, 4, 4), ed3ex(4,4,4), ed5ex(4,4,4),
     &        ehsex(4, 4, 4, 4), eisex(4, 4, 4, 4),
     &        ehex(30), ebex(30), eiex(30),
     &         ei11ex(4,4,4,4,4,4),
     &        ei22aex(4,4,4,4,4,4,4), ei22cex(4,4,4,4,4,4,4),
     &        ei22gex(4,4,4,4,4,4,4), ei22uex(4,4,4,4,4,4,4),
     &        e21sex(4,4,4,4,4,4,4),  e21dex(4,4,4,4,4,4,4),
     &        aex, bex, cex, slopeex,
     &        tphbex, hguex, h3cex, haex, hbex, iasyex, ex3
!
        common /stackex/ esex /dangleex/ ed3ex, ed5ex
     &        /HBIex/ ehex, ebex, eiex
     &        /Hstackex/ ehsex /Istackex/ eisex /int11ex/ ei11ex
     &        /int21ex/ e21sex, e21dex/ int22ex
     &        / ei22aex,  ei22cex, ei22gex, ei22uex
     &        /tetraex/ eh4ex / triex /eh3ex
     &        /other/aex, bex, cex, slopeex,
     &        tphbex, hguex, h3cex, haex, hbex, iasyex, ex3
!
!                set initial values for j-i<=4
!
        do i=1, n
          do j=i, min(i+3, n)
            u(i,j)=pscale(j-i+1)
!
!    2004/09/21 (Clarence):
!    keep up(i,j) as 0.d0 but not -999.d0 here for j<=i+3
!    since recursion later assumes these boundary values
!    to be 0.d0. In other words, the main program should
!    only have set up(i,j) to -999.d0 when i and j are not
!    allowed to pair and j>=i+4
!
            up(i,j)=0.d0
            u1(i,j)=0.d0
            s1(i, j)=0.d0
            s2(i, j)=0.d0
            s3(i, j)=0.d0
          end do
        end do
!
        do i=1, n-4
          termp=1.d0
          if ( augu(i,i+4,n,numseq) ) then
            termp=tphbex
          end if
!
!    2010-05-24 (Ye and Adam) added 3c penalty and GU bonus
!                        3c penalty
          c3p=1.0d0
          if (numseq(i+1) .eq. 2 .and.
     &        numseq(i+2) .eq. 2 .and.
     &        numseq(i+3) .eq. 2) then
            c3p=h3cex
          end if
!
!                       special GU bonus, 5' G preceded by 2 GG &
!            check 2 nt on each side 1st
!            DHM et al say 2 GG in BP, this is not so
!            by mfold,
!                    one case: U followed by C or U for a 3Bp helix
!            but 3rd G can be in closing BP of a bulge!
          gggu=1.0d0
          if ( (i-2) .ge. 1 .and. (i+3+3) .le. n .and.
     &          numseq(i) .eq. 3 .and. numseq(i+3+1) .eq. 4 .and.
     &          numseq(i-1) .eq. 3 .and. numseq(i-2) .eq. 3 ) then
            gggu=hguex
          end if
!
!    2010-07-12 (Adam and Ye) added hairpin indicator (binary) for ehex
!       A+U=5, U+A=5, G+C=5, C+G=5, G+U=7, U+G=7
!       and these are the combinations resulting in sum of 5 or 7
!    2010-08-16 (Adam) removed the above. It is redundant with -999 in main.f
!
!    2004/09/28 (Clarence):
!          bases that cannot be paired will have a up value of
!          -999.d0 which is set in the main program. Leave these
!          values as they are and only initialize all other
!          allowable entries. Include u(i,i+4) in 2 different
!          cases also as u(i,i+4) depends on the value of up(i,i+4)
!          Same treatment for u1(i, i+4) as well.
!
!    2010-05-21 (Ye and Adam) add penalty for AU/GU to up(i, i+4)
!    2010-05-21 (Ye and Adam) add penalty for CCC to
!               up(i, i+4), u(i, i+4), u1(i, i+4)
!    2010-05-24 (Ye and Adam) add penalty for GG preceding GU closing pair to
!               up(i, i+4), u(i, i+4), u1(i, i+4)
!    2010-07-12 (Ye and Adam) add hairpin indicator for ehex to
!               up(i, i+4), u(i, i+4), u1(i, i+4)
!
          if (up(i,i+4) .ne. -999.d0) then
!    2010-05-24 (Ye and Adam) below are the original versions
!         up(i, i+4)=ehex(3)*pscale(5)
!         u(i, i+4)=(1.d0+up(i,i+4)*termp)*pscale(5)
!         u1(i, i+4)=ehex(3)*cex*termp*pscale(5)
!
!    2010-08-19 (Adam) corrected u(i, i+4) to be consistent with
!      up(i, i+4). It can be no other way.
!
!    2010-08-26 (Adam) Because tri-loops energy table does not include
!      terminal penalty, we need to explicitly add it, here. Note that
!      a second termp may also apply due to exterior loop. These are
!      two distinct mechanisms, one for closing base pair of hairpin
!      loop, one for closing base pair of external loop,
!      i.e. not double-counting. See also, s3, below.
!          if ((i .eq. 1) .and. ((i+4) .eq. n)) then
            up(i, i+4)=     (ehex(3)*termp*c3p*gggu) *pscale(5)
            u(i, i+4)=(1.d0+(ehex(3)*termp*c3p*gggu) *termp)*pscale(5)
!    2010-08-26 (Adam) We're not sure about second termp here
            u1(i, i+4)= cex*(ehex(3)*termp*c3p*gggu) *termp *pscale(5)
!          else
!            up(i, i+4)=     ehex(3)          *c3p*gggu *pscale(5)
!            u(i, i+4)=(1.d0+ehex(3)          *c3p*gggu)*pscale(5)
!            u1(i, i+4)=     ehex(3)*cex      *c3p*gggu *pscale(5)
!          end if
!
!     2010-07-13 (Adam): Print each term for debugging
!               write (0,*) "U(",i,", ",i+4,")/pscale(5)=", u(i, i+4)/pscale(5)\
!               write (0,*) "Up",i,", ",i+4,")=", up(i, i+4)
!
          else
            up(i, i+4)=0.d0
            u(i, i+4)=pscale(5)
            u1(i, i+4)=0.d0
          end if
!
          s1(i, i+4)=0.d0
          s2(i, i+4)=0.d0
        end do
!
!                i=1,2 not needed for S3,see p.10 notes 12/7/98
!                    add terminal AU,UA,GU,UG penalty first
        do i=3, n-4
          termp=1.d0
          if ( augu(i,i+4,n,numseq) ) then
            termp=tphbex
          end if
!
!    2010-05-24 (Ye and Adam) added 3c penalty and GU bonus
!                        3c penalty
          c3p=1.0d0
          if (      numseq(i+1) .eq. 2
     & .and. numseq(i+2) .eq. 2
     & .and. numseq(i+3) .eq. 2) then
            c3p=h3cex
          end if
!
!                       special GU bonus, 5' G preceded by 2 GG &
!            check 2 nt on each side 1st
!            DHM et al say 2 GG in BP, this is not so
!            by mfold,
!                    one case: U followed by C or U for a 3Bp helix
!            but 3rd G can be in closing BP of a bulge!
          gggu=1.0d0
          if ( (i-2) .ge. 1 .and. (i+3+3) .le. n .and.
     &         numseq(i) .eq. 3 .and. numseq(i+3+1) .eq. 4 .and.
     &         numseq(i-1) .eq. 3 .and. numseq(i-2) .eq. 3 ) then
            gggu=hguex
          end if
!
!    2010-07-12 (Adam and Ye) added hairpin indicator (binary) for ehex
!       A+U=5, U+A=5, G+C=5, C+G=5, G+U=7, U+G=7
!       and these are the only combinations resulting in sum of 5 or 7.
!    2010-08-16 (Adam) removed the above, as it is already handled in main.f
!
!    2010-05-24 (Ye and Adam) add penalty for CCC to
!               s3(i, i+4)
!    2010-05-24 (Ye and Adam) add penalty for GG preceding GU closing pair to
!               s3(i, i+4)
!    2010-07-12 (Ye and Adam) add hairpin indicator for ehex to
!               s3(i, i+4)
!    2010-07-13 (Ye and Adam) remove mistaken termp from
!               s3(i, i+4)
!               except for special case i=1, i+4=n.
!    2010-07-13 (Ye and Adam) undo previous.
!
!    2004/09/28 (Clarence):
!          split into 2 cases for s3 as well, or otherwise in the
!          multi-loop case in sampling some bases might be sampled
!          but there isn't actually another base that it can pair
!          with, and thus program crashes when the look for the
!          second base goes out of bound
!
!
          if (up(i,i+4) .eq. 0.d0) then
            s3(i, i+4)=0.d0
          else
!    2010-05-24 (Ye and Adam) incorrect version follows
!         s3(i,i+4)=termp*ehex(3)*
!     &        ed5ex(numseq(i),numseq(i+4),numseq(i-1))
!     &        *pscale(5)'
!    2010-07-12 (Ye and Adam) incorrect version follows
!         s3(i,i+4)=termp*c3p*gggu*ehex(3)*
!     &        ed5ex(numseq(i),numseq(i+4),numseq(i-1))
!     &        *pscale(5)
!
!    2010-08-26 (Adam) Not double counting termp, see above.
!    2010-08-26 (Adam) We're not sure about second termp here
            s3(i,i+4)=termp*c3p*gggu*ehex(3)*termp*
     &              ed5ex(numseq(i),numseq(i+4),numseq(i-1))
     &              *pscale(5)
!
!    2010-07-12 (Ye and Adam) retracted, incorrect version follows
!             if ((i .eq. 1) .and. ((i+4) .eq. n)) then
!           s3(i,i+4)=termp*c3p*gggu*hp*ehex(3)*
!     &          ed5ex(numseq(i),numseq(i+4),numseq(i-1))
!     &          *pscale(5)
!             else
!           s3(i,i+4)=c3p*gggu*hp*ehex(3)*
!     &          ed5ex(numseq(i),numseq(i+4),numseq(i-1))
!     &          *pscale(5)
!             end if
          end if
        end do
!
!                special boundary values
        do i=1, n
          u(i+1, i)=1.d0
          u1(i+1, i)=0.d0
        end do
!
        do i=1, n-1
          u1(i+2, i)=0.d0
        end do
!
!                compute rest of up and associated PFs u1, s2, s3
!                start with increment (inc) j-i=4
!
!    2004/12/20 (Clarence):
!    as suggested by Ye, changes were made to big loop indices
!    to reduce cputime from n^3 to n(lflen^2) where lflen is the
!    length of local folding
!
!    do 10 inc=4, n-2
        do 10 inc=4, min(lflen-1, n-2)
          do 20 i=1, n-inc-1
            if ((i+inc+1) .gt. n) then
              go to 20
            end if
!            compute s2(i, i+inc+1), not i=1,2,3,p9 notes 12/7/98
            if (i .le. 3) then
              go to 15
            end if
!
            s2(i, i+inc+1)=0.d0
!
            do l=i+1, i+inc
!                    add terminal AU,UA,GU,UG penalty
              termp=1.d0
              if ( augu(i,l,n,numseq) ) then
                termp=tphbex
              end if
!
!    2004/09/21 (Clarence):
!       Again, set s2() in 2 different cases since s2(i, i+inc+1)
!       depends on the value of up(i,l), which can be the special
!       value of -999.d0 in some cases. In those cases, copy the
!       original equation and set up(i,l) to 0.d0 in it as the
!       equation is long and we don't want to mess it up with the
!       wrong simplified equation
!
              if (up(i,l) .ne. -999.d0) then
                s2(i, i+inc+1)=s2(i, i+inc+1)+
     &            up(i,l)*termp*
     &            ed5ex(numseq(i),numseq(l),numseq(i-1))*
     &            (  ed3ex(numseq(i),numseq(l),numseq(l+1))
     &            *u1(l+2, i+inc)*pscale(1)
     &            +u1(l+1, i+inc)-u1(l+2, i+inc)*pscale(1)  )
     &            *pscale(1)
              else
                s2(i, i+inc+1)=s2(i, i+inc+1)+
     &            0.d0*termp*
     &            ed5ex(numseq(i),numseq(l),numseq(i-1))*
     &            (  ed3ex(numseq(i),numseq(l),numseq(l+1))
     &            *u1(l+2, i+inc)*pscale(1)
     &            +u1(l+1, i+inc)-u1(l+2, i+inc)*pscale(1)  )
     &            *pscale(1)
              end if
!
            end do
!
15          continue
!
!                    compute up(i, i+inc+1)
!                    by uph, ups, upbr, upbl, upi, upm first
!                        uph(i, i+inc+1) for hairpin
!
!    2004/09/21 (Clarence):
!    We start to split the case for up() here. if up(i, i+inc+1)
!    is preset to -999.d0, we know that bases i and i+inc+1 are
!    not allowed to pair, so we skip the whole calculation of
!    up(i, i+inc+1) and just simply set it to 0.d0. For the other
!    case where up(i, i+inc+1) is not -999.d0, we go through the
!    original steps in computing the value. I have double-checked
!    that all references to up(i',j') in this block of codes are
!    for i<=i'<j'<=j, which should guarantee only previous up
!    values are used and so no -999.d0 should be included in the
!    real calculation.
!
            if (up(i, i+inc+1) .eq. -999.d0) then
              up(i, i+inc+1) = 0.d0
              go to 11
            end if
!
!     2010-08-16 (Adam) Refactored hpe calculation. Use hpeex() function.
            uph=hpeex(i, i+inc+1, n, numseq, ehex, eh4ex, ehsex,
     &               slopeex, hguex, h3cex, haex, hbex, tphbex)*pscale(inc+2)
!
!                    ups(i, i+inc+1) for stacking
!
            ups=up(i+1, i+inc)*
     &             esex(numseq(i), numseq(i+1),
     &                numseq(i+inc), numseq(i+inc+1))
            ups=ups*pscale(2)
!
!                        upb(i, i+inc+1) for bulges
!                l starts @ i+5, for >= 3nt for hairpin (i+1)--l
            upbr=0.d0
            do l=i+5, i+inc-1
!                add stacking if size=1, to loop penalty
!                or au,gu penalty if size >= 2
              prod1=1.d0
              prod2=1.d0
              if ( augu(i,i+inc+1,n,numseq) )  then
                prod1=tphbex
              end if
              if ( augu(l,i+1,n,numseq) ) then
                prod2=tphbex
              end if
!
              if ((i+inc-l) .gt. 30) then
                upbr=upbr+prod1*prod2*
     &             ebex(30)*( slopeex**dlog(dble(i+inc-l)/30.d0) )
!     &         ebex(30)*dexp( -(1000.d0/(1.9872d0*310.15d0))
!     &            *1.079d0*dlog(dble(i+inc-l)/30.d0) )
     &            *up(i+1,l)*pscale(i+inc+2-l)
!
              else
                if (l .eq. (i+inc-1)) then
                  upbr=upbr+ebex(1)*up(i+1,l)*
     &              esex(numseq(i), numseq(i+1),
     &                numseq(i+inc-1), numseq(i+inc+1))
     &              *pscale(i+inc+2-l)
                else
                  upbr=upbr+prod1*prod2*ebex(i+inc-l)
     &                *up(i+1,l)
     &              *pscale(i+inc+2-l)
                end if
              end if
            end do
!
            upbl=0.d0
!              h ends @ i+inc-4, for >= 3nt for h--(i+inc) hairpin
            do h=i+2,  i+inc-4
!               add stacking if size=1, or au,gu penalty if size >= 2
              prod1=1.d0
              prod2=1.d0
              if (augu(i,i+inc+1,n,numseq)) then
                prod1=tphbex
              end if
              if  ( augu(i+inc,h,n,numseq) ) then
                prod2=tphbex
              end if
!
              if ((h-i-1) .gt. 30) then
                upbl=upbl+prod1*prod2*
     &            ebex(30)*( slopeex**dlog(dble(h-i-1)/30.d0) )
!     &          ebex(30)*dexp( -(1000.d0/(1.9872d0*310.15d0))
!     &            *1.079d0*dlog(dble(i+inc-l)/30.d0) )
     &            *up(h,i+inc)*pscale(h-i+1)
              else
                if (h .eq. (i+2)) then
                  upbl=upbl+ebex(h-i-1)*up(h,i+inc)*
     &                       esex(numseq(i), numseq(i+2),
     &                numseq(i+inc), numseq(i+inc+1))
     &             *pscale(h-i+1)
                else
                  upbl=upbl+prod1*prod2*ebex(h-i-1)
     &            *up(h,i+inc)
     &            *pscale(h-i+1)
                end if
              end if
            end do
!
!                              upi(i, i+inc+1) for interior
            upi=0.d0
            do h=i+2, min(i+30, i+inc-5)
              do l=max(h+4, h+inc-31), i+inc-1
!
!                    1x1 I loop, use 1x1 table
                if(h .eq. (i+2) .and. l .eq. (i+inc-1)) then
                  upi=upi+
     &              ei11ex(numseq(i), numseq(i+inc+1),
     &                numseq(i+2), numseq(i+inc-1),
     &                numseq(i+1), numseq(i+inc))*up(h,l)
     &              *pscale(h+inc+1-l)
                  go to 6
                end if
!
!                1x2 loop, single nt on 5' --> 3'
                if (h .eq. (i+2) .and. l .eq. (i+inc-2)) then
                  upi=upi+
     &              e21sex(numseq(i), numseq(i+inc+1),
     &              numseq(i+2), numseq(i+inc-2),
     &              numseq(i+1),
     &              numseq(i+inc-1), numseq(i+inc))*up(h,l)
     &              *pscale(h+inc+1-l)
                  go to 6
                end if
!
!                2x1 loop, double nt on 5' --> 3'
                if (h .eq. (i+3) .and. l .eq. (i+inc-1)) then
                  upi=upi+
     &              e21dex(numseq(i), numseq(i+inc+1),
     &              numseq(i+3), numseq(i+inc-1),
     &              numseq(i+1), numseq(i+2),
     &              numseq(i+inc))*up(h,l)
     &              *pscale(h+inc+1-l)
                  go to 6
                end if
!
!                    2x2 loop
                if (h .eq. (i+3) .and. l .eq. (i+inc-2)) then
                  if (numseq(i+inc) .eq. 1) then
                    upi=upi+
     &                ei22aex(numseq(i), numseq(i+inc+1),
     &                numseq(i+3), numseq(i+inc-2),
     &                numseq(i+1), numseq(i+2),
     &                numseq(i+inc-1))*up(h,l)
     &                *pscale(h+inc+1-l)
                  end if
!
                  if (numseq(i+inc) .eq. 2) then
                    upi=upi+
     &                ei22cex(numseq(i), numseq(i+inc+1),
     &                numseq(i+3), numseq(i+inc-2),
     &                numseq(i+1), numseq(i+2),
     &                numseq(i+inc-1))*up(h,l)
     &                *pscale(h+inc+1-l)
                  end if
!
                  if (numseq(i+inc) .eq. 3) then
                    upi=upi+
     &                ei22gex(numseq(i), numseq(i+inc+1),
     &                numseq(i+3), numseq(i+inc-2),
     &                numseq(i+1), numseq(i+2),
     &                numseq(i+inc-1))*up(h,l)
     &                *pscale(h+inc+1-l)
                  end if
!
                  if (numseq(i+inc) .eq. 4) then
                    upi=upi+
     &                ei22uex(numseq(i), numseq(i+inc+1),
     &                numseq(i+3), numseq(i+inc-2),
     &                numseq(i+1), numseq(i+2),
     &                numseq(i+inc-1))*up(h,l)
     &                *pscale(h+inc+1-l)
                  end if
!
                  go to 6
                end if
!
!                all other cases, > 4 nt and 1x3, 3x1 (4 nt)
                if ( ((h+inc-l-1) .gt. 4 ) .or.
     &                (h .eq. (i+2) .and. l .eq. (i+inc-3)) .or.
     &               ( h .eq. (i+4) .and. l .eq. (i+inc-1))  ) then
!
!                2 terminal stacks +asym. penalty,maximum is 3
!        this limit is used by mfold, but not made clear by DHM
!                   for 1Xn,nX1 treat 1 nt as 2 same nt(DHM email 1/7/99)
!               and treat the 3 nts involved in 2 mismatches as A
!               (DHM email 2/8/99) i.e, take numseq(*)=1
                  if (h .eq. (i+2) .or. l .eq. (i+inc-1)) then
                    prod1=eisex(numseq(i), 1, 1, numseq(i+inc+1))
     &              *dmax1(ex3, iasyex**abs(dble((h-i-1)-(i+inc-l))) )
!        dmax1 gives same result as min for the original=cc
!     &      *dexp(-(1000.d0/(1.9872d0*310.15d0))
!     &        *min(3.d0, 0.5d0*abs(dble((h-i-1)-(i+inc-l)))))
!!   &      *dexp(-scale*min(3.d0, iasy*abs(dble((h-i-1)-(i+inc-l)))))
     &              *eisex(numseq(l), 1, 1, numseq(h))
!
                  else
!
                    prod1=eisex(numseq(i),
     &                numseq(i+1), numseq(i+inc), numseq(i+inc+1))
     &                *dmax1(ex3, iasyex**abs(dble((h-i-1)-(i+inc-l))) )
!        dmax1 gives same result as min forthe original=cc
!     &      *dexp(-(1000.d0/(1.9872d0*310.15d0))
!     &        *min(3.d0, 0.5d0*abs(dble((h-i-1)-(i+inc-l)))))
!!   &    *dexp(-scale*min(3.d0, iasy*abs(dble((h-i-1)-(i+inc-l)))))
     &                *eisex(numseq(l), numseq(l+1),
     &                 numseq(h-1), numseq(h))
                  end if
                  upi=upi+eiex(h+inc-l-1)*up(h,l)
     &               *prod1*pscale(h+inc+1-l)
!
                end if
!
6               continue
!
              end do
            end do
!
!                    compute upm
            sum1=0.d0
            do l=i+2, i+inc
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i+1,l,n,numseq) )  then
                termp=tphbex
              end if
!
              sum1=sum1+up(i+1,l)*termp*
     &           (  ed3ex(numseq(i+1),numseq(l),numseq(l+1))
     &                *u1(l+2, i+inc)*pscale(1)
     &           +u1(l+1, i+inc)-u1(l+2, i+inc)*pscale(1)  )
!
            end do
!           2010-10-20 (Adam) Factored out common terms in large sum.
            sum1=sum1*pscale(2)
!
            sum2=0.d0
            do l=i+3, i+inc
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i+2,l,n,numseq) )  then
                termp=tphbex
              end if
!
!                5's single stack is ignored but included in c.
              sum2=sum2+
!     &  dmax1(ed3ex(numseq(i+inc+1), numseq(i), numseq(i+1),
!     &            ed5ex(numseq(i+2),numseq(l),numseq(i+1))))
     &          ed3ex(numseq(i+inc+1), numseq(i), numseq(i+1))
     &          *up(i+2,l)*termp*
     &          ( ed3ex(numseq(i+2),numseq(l),numseq(l+1))
     &          *u1(l+2, i+inc)*pscale(1)
     &          +u1(l+1, i+inc)-u1(l+2, i+inc)*pscale(1)  )
!
            end do
!           2010-10-20 (Adam) Factored out common terms in large sum.
            sum2=sum2*bex*pscale(3)
!
            sum3=0.d0
            do h=i+3, i+inc-1
              sum3=sum3+pscale(h-i)*(bex**dble(h-i-1))*s2(h, i+inc+1)
            end do
!           2010-10-20 (Adam) Factored out common terms in large sum.
            sum3=sum3*ed3ex(numseq(i+inc+1), numseq(i),numseq(i+1))
!
!     2010-09-24 (Adam&Ye) termp on the upm term as well (not in paper A2)
            termp=1.d0
            if ( augu(i,i+inc+1,n,numseq) )  then
              termp=tphbex
            end if
!           2010-10-20 (Adam) Factored out common terms in large sum.
            upm=aex*(cex**2.d0)*(sum1+sum2+sum3)*termp
!
            up(i, i+inc+1)=uph+ups+upbr+upbl+upi+upm
!
!    Print the components of up(i, i+inc+1) for debugging
!            if (i .eq. 28 .and. i+inc+1 .eq. 33) then
!              write (0,*) "Up(",i,", ",i+inc+1,")=Uph+Ups+Upbr+Upbl+Upi+Upm"
!              write (0,*) "              =",up(i, i+inc+1)
!              write (0,*) "uph=",uph
!              write (0,*) "ups=",ups
!              write (0,*) "upbr=",upbr
!              write (0,*) "upbl=",upbl
!              write (0,*) "upi=",upi
!              write (0,*) "upm=",upm
!            end if
!
!    2004/09/21 (Clarence):
!    this is the end of the if-else case for
!    (up(i, i+inc+1) .eq. -999.d0)
!
11          continue
!
!                    compute u1(i, i+inc+1)
            sum1=0.d0
            do l=i+1, i+inc+1
!
             stack3=1.d0
              if (l .eq. (i+inc+1)) then
!               leave stack3=1.0d
!               2011-01-26 Adam: the following go to appears to have no purpose.
                go to 7
              end if
!            in commented out parts of next 2 ifs, ed5(*,-) for 5'
!            stack on 3' end of closing BP of M-loop,
!            * to be simulated or average used, it is ignored now
!
              if (l .eq. (i+inc)) then
                stack3=ed3ex(numseq(i), numseq(l),numseq(l+1))
!      stack3=dmax1(ed3ex(numseq(i),numseq(l),numseq(l+1)),
!      &    ed5ex(numseq(*),numseq(i+inc+2), numseq(i+inc+1))))
                go to 7
              end if
!
              if (l .le. (i+inc-1)) then
                stack3=ed3ex(numseq(i), numseq(l),numseq(l+1))
!     &        ed3ex(numseq(i),numseq(l),numseq(l+1))
!     &           *ed5ex(numseq(*),numseq(i+inc+2), numseq(i+inc+1))
              end if
!
7             continue
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i,l,n,numseq) ) then
                termp=tphbex
              end if
!
              if (l .eq. (i+inc+1)) then
                scaleterm=1.d0
              else
                scaleterm=pscale(i+inc+1-l)
              end if
!
              if ((l+1) .gt. n) then
                tmp=up(i,l)*termp*
     &               ( stack3*(bex**dble(i+inc+1-l))
     &            *scaleterm )
              else
                tmp=up(i,l)*termp*
     &               ( stack3*(bex**dble(i+inc+1-l))
     &            *scaleterm+
     &            ed3ex(numseq(i),numseq(l),numseq(l+1))
     &            *u1(l+2, i+inc+1)*pscale(1)+
     &              u1(l+1, i+inc+1)-u1(l+2, i+inc+1)*pscale(1) )
              end if
              sum1=sum1+tmp
            end do
!
            sum2=0.d0
            do l=i+2, i+inc+1
!
              stack3=1.d0
              if (l .eq. (i+inc+1)) then
!               2011-01-26 Adam: the following goto statement appears to me to be extraneous
                go to 8
              end if
!
!            in commented out parts of next 2 ifs, ed5(*,-) for 5'
!            stack on 3' end of closing BP of M-loop,
!            * to be simulated or average used, it is ignored now
!
              if (l .eq. (i+inc)) then
                stack3=ed3ex(numseq(i+1), numseq(l),numseq(l+1))
!    stack3=dmax1(ed3ex(numseq(i+1),numseq(l),numseq(l+1)),
!     &        ed5ex(numseq(*),numseq(i+inc+2), numseq(i+inc+1))
                go to 8
              end if
!
              if (l .le. (i+inc-1)) then
                stack3=ed3ex(numseq(i+1), numseq(l),numseq(l+1))
!     &        ed3ex(numseq(i+1),numseq(l),numseq(l+1))
!     &        *ed5ex(numseq(*),numseq(i+inc+2), numseq(i+inc+1))
              end if
!
8             continue
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i+1,l,n,numseq) ) then
                termp=tphbex
              end if
!
              if (l .eq. (i+inc+1)) then
                scaleterm=1.d0
              else
                scaleterm=pscale(i+inc+1-l)
              end if
!
              if((l+1) .gt. n) then
                tmp=up(i+1,l)*termp*
     &               ( stack3*(bex**dble(i+inc+1-l))
     &            *scaleterm )
              else
                tmp=up(i+1,l)*termp*
     &               ( stack3*(bex**dble(i+inc+1-l))
     &             *scaleterm+
     &            ed3ex(numseq(i+1),numseq(l),numseq(l+1))
     &            *u1(l+2, i+inc+1)*pscale(1)+
     &              u1(l+1, i+inc+1)-u1(l+2, i+inc+1)*pscale(1) )
              end if
              sum2=sum2+tmp
!
            end do
!           2010-10-20 (Adam) Factored out common terms in large sum.
            sum2=sum2*bex*pscale(1)
!
            sum3=0.d0
            do h=i+2, i+inc
              tmp=pscale(h-i)*(bex**dble(h-i))*s3(h, i+inc+1)
              sum3=sum3+tmp
            end do
!
!           2010-10-20 (Adam) Factored out common terms in large sum.
            u1(i, i+inc+1)=cex*(sum1+sum2+sum3)
!
!                    compute s3(i, i+inc+1)
!                    not i=1,2, see p10., notes 12/7/98
            if (i .le. 2) then
              go to 20
            end if
!
            s3(i, i+inc+1)=0.d0
            do l=i+1, i+inc+1
!
              stack3=1.d0
              if (l .eq. (i+inc+1)) then
!               2011-01-26 Adam: The following goto statement appears to me to be extraneous.
                go to 9
              end if
!
!            in commented out parts of next 2 ifs, ed5(*,-) for 5'
!            stack on 3' end of closing BP of M-loop,
!            * to be simulated or average used, it is ignored now
!
              if (l .eq. (i+inc)) then
                stack3=ed3ex(numseq(i),numseq(l),numseq(l+1))
!      stack3=dmax1( ed3ex(numseq(i),numseq(l),numseq(l+1)),
!     &            ed5ex(numseq(*),numseq(i+inc+2), numseq(i+inc+1)) )
                go to 9
              end if
!
              if (l .le. (i+inc-1)) then
                stack3=ed3ex(numseq(i), numseq(l),numseq(l+1))
!     &          ed3ex(numseq(i),numseq(l),numseq(l+1))
!     &          *ed5ex(numseq(*),numseq(i+inc+2), numseq(i+inc+1))
              end if
!
9             continue
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i,l,n,numseq) ) then
                termp=tphbex
              end if
!
              if (l .eq. (i+inc+1)) then
                scaleterm=1.d0
              else
                scaleterm=pscale(i+inc+1-l)
              end if
!
              if ((l+1) .gt. n) then
                s3(i, i+inc+1)=s3(i, i+inc+1)+
     &            up(i, l)*termp*
     &            ed5ex(numseq(i), numseq(l), numseq(i-1))*
     &            (  stack3*(bex**dble(i+inc+1-l))
     &                *scaleterm )
              else
                s3(i, i+inc+1)=s3(i, i+inc+1)+
     &            up(i, l)*termp*
     &            ed5ex(numseq(i), numseq(l), numseq(i-1))*
     &            (  stack3*( bex**dble(i+inc+1-l) )
     &                *scaleterm+
     &            ed3ex(numseq(i),numseq(l),numseq(l+1))
     &                *u1(l+2, i+inc+1)*pscale(1)+
     &              u1(l+1, i+inc+1)-u1(l+2, i+inc+1)*pscale(1) )
              end if
            end do
!
20        continue
10      continue
!
!                    now compute s1(i,j), u(i,j), j-i>5
!
!    2004/12/20 (Clarence):
!    WARNING: the following loop indices should NOT be changed to
!    accomodate the introduction of lflen to partf.f. This is
!    because the following code, of O(n^3), is used to compute
!    s1(i,j) and u(i,j). They cannot be set to zero simply as they
!    are used later in sampling to compute the probability, not
!    even for (i,j) outside the local folding limit as we start
!    sampling from (1,n). does not seem to have an alternative way
!    to set the correct values with smaller computational order.
!
        do inc=4, n-2
!    do inc=4, min(lflen-1, n-2)
          do i=1, n-inc-1
!
!                compute s1(i, i+inc+1), not i=1, p9 notes 12/7/98
            if (i .eq. 1) then
              go to 30
            end if
            s1(i,i+inc+1)=0.d0
            do l=i+1, i+inc
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i,l,n,numseq) ) then
                termp=tphbex
              end if
!
              s1(i,i+inc+1)=s1(i,i+inc+1)+
     &            up(i, l)*termp
!     2010-09-13 (Adam) Removed ed5 here becuase we don't know if i-1 is avail
!     2010-09-13 (Adam) Removed ed3 here becuase we don't know if l+1 is avail
     &                *u(l+1, i+inc+1)
            end do
!
30          continue
!                    compute u(i, i+inc+1)
            sum1=0.d0
            do l=i+1, i+inc
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i,l,n,numseq) ) then
                termp=tphbex
              end if
!
              sum1=sum1 + up(i, l)*termp
!     2010-09-13 (Adam) Removed ed3 here because we don't know if l+1 is avail
     &          *u(l+1, i+inc+1)
!
            end do
!
            sum2=0.d0
            do h=i+1, i+inc
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(h,i+inc+1,n,numseq) ) then
                termp=tphbex
              end if
!
              sum2=sum2 + up(h,i+inc+1)*termp *pscale(h-i)
!     2010-09-13 (Adam) Removed ed5 here becuase we don't know if h-1 is avail
!     &        *ed5ex(numseq(h), numseq(i+inc+1), numseq(h-1))
            end do
!
            sum3=0.d0
            do h=i+1, i+inc-1
              sum3=sum3+s1(h,i+inc+1)*pscale(h-i)
            end do
!
!                 add au/gu terminal penalty
            termp=1.d0
            if ( augu(i,i+inc+1,n,numseq) ) then
              termp=tphbex
            end if
!
            u(i, i+inc+1)=pscale(inc+2)
     &        + up(i, i+inc+1)*termp
     &        + sum1+sum2+sum3
!
!     2010-07-13 (Adam): Print each term for debugging
!               write (0,*) "U(",i,", ",i+inc+1,")/pscale(", (inc+2), ")=", u(i, i+inc+1)/pscale(inc+2)
!               write (0,*) " pscale(",inc+2,")=", pscale(inc+2)
!               write (0,*) " Uphp + (Up(",i,", ",i+inc+1,") - Uphp)*termp=",
!     &          uphp + (up(i, i+inc+1) - uphp)*termp
!               write (0,*) " termp=", termp
!               write (0,*) " sum1=", sum1
!
!          do l=i+1, i+inc
!         termp=1.d0
!         if ( ((numseq(i) .eq. 1) .and. (numseq(l) .eq. 4)) .or.
!     &              ((numseq(i) .eq. 4) .and. (numseq(l) .eq. 1)) .or.
!     &              ((numseq(i) .eq. 3) .and. (numseq(l) .eq. 4)) .or.
!     &              ((numseq(i) .eq. 4) .and. (numseq(l) .eq. 3)) ) then
!                 termp=tphbex
!              end if
!
!                 write (0,*) "  sum1_",l,"=Up(i, l)*termp*",
!     &          "( ed3ex(numseq(i),numseq(l),numseq(l+1))*",
!     &            "U(l+2, i+inc+1)*pscale(1)+",
!     &              "U(l+1, i+inc+1)-U(l+2, i+inc+1)*pscale(1) )"
!                 write (0,*) "  sum1_",l,"=", up(i, l)*termp*
!     &          ( ed3ex(numseq(i),numseq(l),numseq(l+1))*
!     &            u(l+2, i+inc+1)*pscale(1)+
!     &              u(l+1, i+inc+1)-u(l+2, i+inc+1)*pscale(1) )
!                 write (0,*) "   Up(",i,", ",l,")=", up(i, l)
!                 write (0,*) "   termp=", termp
!                 write (0,*) "   ed3ex(", numseq(i), ", ", numseq(l), ", ",
!     &                      numseq(l+1), ")=",
!     &                      ed3ex(numseq(i), numseq(l), numseq(l+1))
!                 write (0,*) "   U(",l+2,", ",i+inc+1,")=", u(l+2, i+inc+1)
!                 write (0,*) "   pscale(1)=", pscale(1)
!                 write (0,*) "   U(",l+1,", ",i+inc+1,")=", u(l+1, i+inc+1)
!                 write (0,*) "   U(",l+2,", ",i+inc+1,")=", u(l+2, i+inc+1)
!              end do
!
!               write (0,*) " sum2=", sum2
!
!          do h=i+1, i+inc
!
!        termp=1.d0
!        if ( ((numseq(h) .eq. 1) .and. (numseq(i+inc+1) .eq. 4)) .or.
!     &             ((numseq(h) .eq. 4) .and. (numseq(i+inc+1) .eq. 1)) .or.
!     &             ((numseq(h) .eq. 3) .and. (numseq(i+inc+1) .eq. 4)) .or.
!     &             ((numseq(h) .eq. 4) .and. (numseq(i+inc+1) .eq. 3)) ) then
!               termp=tphbex
!                end if
!
!                 write (0,*) "  sum2_",h,"=", pscale(h-i)*up(h, i+inc+1)*termp*
!     &        ed5ex(numseq(h), numseq(i+inc+1), numseq(h-1))
!                 write (0,*) "   pscale(",h-i,")=", pscale(h-i)
!                 write (0,*) "   Up(",h,", ",i+inc+1,")=", up(h, i+inc+1)
!                 write (0,*) "   termp=", termp
!                 write (0,*) "   ed5ex(", numseq(h), ", ", numseq(i+inc+1), ", ",
!     &                      numseq(h-1), ")=",
!     &                      ed5ex(numseq(h), numseq(i+inc+1), numseq(h-1))
!              end do
!
!               write (0,*) " sum3=", sum3
!
!          do h=i+1, i+inc-1
!                 write (0,*) "  sum3_",h,"= S1(h,i+inc+1)*pscale(h-i)"
!                 write (0,*) "  sum3_",h,"=", s1(h,i+inc+1)*pscale(h-i)
!                 write (0,*) "   S1(",h,", ",i+inc+1,")=", s1(h, i+inc+1)
!                 write (0,*) "   pscale(",h-i,")=", pscale(h-i)
!              end do
!
!               u(i, i+inc+1)=pscale(inc+2)+up(i, i+inc+1)+sum1+sum2+sum3
          end do
        end do
!
!     2010-06-08 (Adam and Ye): Verify u(i, j)-u(i+1, j) > 0
!        do i=1, n+1
!          do j=i-1, n
!             write (0,*) "U(", i,   ", ", j, ")=", u(i, j)
!             write (0,*) "Up(", i,   ", ", j, ")=", up(i, j)
!            if (u(i, j)-u(i+1, j)*pscale(1) .lt. 0.0d0) then
!              write (0,*) "i=", i, ",j=",  j, " length=", j-i+1, " error found in U()."
!              write (0,*) "U(", i,   ", ", j, ")          =", u(i, j)
!              write (0,*) "U(", i+1, ", ", j, ")*pscale(1)=" , u(i+1, j)*pscale(1)
!            end if
!          end do
!        end do
!
        return
      end
