!!
!! pprob.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!  This is the f77 subroutine to compute the
!  (a) minimum free energy of sampled structures (Smfe),
!  (b) the Bolztman probability of Smfe,
!  (c)  the estimated probability of within p%
!    of this minimum, p=0, ...100 ("CDF")
!  (d) and the probability for each P interval of witch 5 ("PDF"),
!
!  (e) the option of mannually inputing mfe from mfold, then
!    (b), (c) and (d) are computed for mfe, insteda of Smfe
!  Note 1: the minimum is usually the MFE for shorter seq (100+ bases)
!    however, it is often not for long seq(400), bacause too many
!    near optimal structures with similar FE
!  Note2:in a collabaration with Zuker, can use mfe by mfold to replace
!    Smfe, however. for 16SEcoli D2, Smfe < mfe, partly because
!    Sfold permit internal helx w/o 1 bp,when I input Smfe structure
!    mfold does ouput same lower energy! this needs to
!    straightened out, before using mfe, or we report
!    min(Smfe, MCP ): MCP: maximum conditional prob.
!  Note 3: p=100 corresp. to e=0, there are cases for short RNA like Sllc
!  sampled energy is > 0 for 1 or 2 cases among 1,000.
!  Thus p form 0 to 100 does not consider such rare cases

!    2006/10/13 (Clarence):
!       Function paramter list updated to include new variables
!  outdir and outdirlen for the location to write output
!  files. See main.f for more details.
      subroutine pprob(n,nsample, u0, u,fe, pnum, probp, rightpt, probint,
     &       outdir, outdirlen)
        integer  nsample,  pnum(101), counter, rightpt(21),n,i,j
        double precision probp(101), fe(nsample), minfe, febound,
     &       probint(21), feboundl, u0,u(n+1, n+1), BMprob, scale
        character*(*) outdir
        integer outdirlen

!    2006/11/16 (Clarence): updated the way to pass filename to open().
!       both intel compiler and sun compiler take formats like
!       open(unit=2, file=(outdir(1:outdirlen) // '/sample.out'), but
!       gcc would fail for the filename part with the concatenation.
!       So we need to assign the filename to a variable first, then
!       pass it along to open(). bufstr would be the temp variable for
!       storing filename and buflen is its max length.
!
        integer buflen
        parameter(buflen=1000)
        character*(buflen) bufstr

!        compute the min. free energy of sampled
        bufstr = outdir(1:outdirlen) // '/smfe.out'
        open(unit=11, file=bufstr(1:len_trim(bufstr)),
     &         status='replace')

        minfe=fe(1)
        do i=2, nsample
          if( fe(i) .lt. minfe) then
            minfe=fe(i)
          end if
        end do

!    here mannualy input Zuker's mfe, comment out if not
!    -10.7 for sllc, -28.9 for tRNA-ala, -56.2 for 5S Xlo
!    -48.5 for CIII,  -151.3 (-152.5 by Smfe) for  16SEcoli D2
!    -173.6 for Ecoli RNase P. -149.3 (2 structure) for TT inton 1
!
!  minfe=-149.3
!  minfe=-173.6
!        compute Bolztmann probability for minfe
        scale=1000.d0/(1.9872*310.15)
!        unscaled u(1,n)=scaled u(1,n)*(u0**n), see notes.
!        Need to avoid overflow here by using log!
        BMprob=dexp(-scale*minfe-( dlog(u(1,n))+dble(n)*dlog(u0) ) )
        write (11, 10)  minfe
        write (11, 20)   BMprob
!  write (11, 30)   u(1,n)
 10     format('Minimum free energy in the sample (SMFE)=',
     &            f8.2, 1x,'kcal/mol')
 20     format('Boltzmann probability of SMFE structure =',  e15.5)
!30  format('u(1,n) =',  e15.5)
!  print *,  u(1,n)
!      compute est. prob. <= (1-p/100) minfe for each p
        do i=1, 101

          pnum(i)=i-1
          febound=(1.d0-dble(pnum(i))/100.d0)*minfe
          counter=0

          do j=1, nsample
            if(fe(j) .le. febound ) then
              counter=counter+1
            end if
          end do

          probp(i)=dble(counter)/dble(nsample)
        end do

!        compute est. prob. for energy intervals ,
!        width=5, 5*(i-2) < int <=5*(i-1), i=1,... 21
!        i=1 for prob at sampled FE
        do i=1, 21
          counter=0
          feboundl=(1.d0-(5.0*dble(i-2))/100.d0)*minfe
          febound=(1.d0-(5.0*dble(i-1))/100.d0)*minfe

          do j=1, nsample
            if (fe(j) .gt. feboundl .and. fe(j) .le. febound) then
              counter=counter+1
            end if
          end do
          probint(i)=dble(counter)/dble(nsample)

        end do

        do i=1, 21
          rightpt(i)=5*(i-1)
        end do

        close(11)

        return
      end
