!!
!! sample.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!    compaq version, use 4 7D arrays ei22a,ei22ac,ei22ag, ei22aU
!!!!    Rescale version to avoid overflow, u0**n is est or exact u(1,n)
!!        pscale(i)=1/(u0**i), =0 for large i (underflow)
!!  use Mersenne Twister u(0,1) for discrete sampling
!!!!!This is  the subroutine to sample a base pair (or base pairs for M-loop)
!!!!!single base involving the rightmost ribonucleotide in any subsequence,
!!!!! returning single bases from H, B, I and M loop
!!!!!6 is for 6 cases, H,S, Br, BL, I and M-loop
!    sn: starting base, en: ending base
!    ind=1, if sn.en known, and o otherwise
!    seq0(2): return active subsequnce with ind=0;1:start, 2:end
!    seq1(n:2): return active subsequence (sequences for M-loop) with ind=1
!    bp(n:2): return base pair from H, S, B & I, base pairs for M loop
!    single(n): single base or bases
!    ltype(n): loop type of unpaires base(s): 1-H, 2-B, 3-I, 4-M
!          5-ext; 0-base paired (initials)
!    note: 0 for seq0, seq1, bp, single indicate no new subsequnce/bp/single
!    probm(nm): holds discrete probabilities, although size varies, but
!           drngda allows 0 entries, upper bound for nm is set 2+(n-1)*t
!    input:  update for each sample: sn, en, ind
!    output: seq0,seq1, bp, single, ltype
!    Note 1: the output is for sampling for 1-fragment, thus output need
!          to be put together in main program for a complete structure !
!    Note 2: h-loops with en-sn < 6 with sn.en pair too short for sampling
!        this is handled in main program !
!    helixn: # of helices-1 for M-loop, hm(n)-lh(n): BPs on M-loop
!    for I-loop sampling, probmi(450) is ok for 30*29/2 < 450,
!            for size <=30, n**2/2 array needed for no size limit
!
! 2004/09/28 (Clarence): replace all 'print *' statement to 'write (0, *)'.
!    Instead of printing to standard output, 'print *' seems to print to
!    some other open file handle...
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine sample (n, numseq, sn, en, ind, pscale,
     &         u, up,  u1,
     &        s1, s2, s3, ir6, hm, lm,
     &        probmbr, probmbl, probm1, probm2,
     &               probmm,    seq0, seq1, bp, single, ltype)
!
        integer n, numseq(n), sn, en, ind, i, inc
        integer seq0(2), seq1(n,2), bp(2), single(n), ltype(n)
        integer nr, nb, n6, ni
        integer h, l, k, h0, hm(n), lm(n), helixn, i0, j0
        double precision hpeex
        logical augu, do_pre_normalize_check, binsamp_de, binsamp_mbl
!
        parameter(nr=1, nb=2, n6=6, ni=450)
        integer
     &    ir(nr), ir2(nr), ir6(nr), irbr(nr), irbl(nr),
     &  iri(nr), irm(nr), irbn(nr), irbnss1(nr)
!
        double precision pscale(n),
     &        u(n+1, n+1),up(n, n), uph, ups, upbr, upbl,
     &        upi, upm, sum1, sum2, sum3, u1(n+1, n+1),
     &        s1(n,n),s2(n,n),s3(n,n),  probm6(n6),
     &         probm1(3*n),
     &        probm2(n), probmbr(n), probmbl(n), probmi(ni),
     &         probmm(3*n),  probb(nb),
     &        prod1, prod2, stack3, sump, scaleterm
!
        double precision  eh3ex(4,4,4,4,4), eh4ex(4,4,4,4,4,4),
     &         esex(4, 4, 4, 4), ed3ex(4,4,4), ed5ex(4,4,4),
     &        ehsex(4, 4, 4, 4), eisex(4, 4, 4, 4),
     &        ehex(30), ebex(30), eiex(30),
     &         ei11ex(4,4,4,4,4,4),
     &        ei22aex(4,4,4,4,4,4,4), ei22cex(4,4,4,4,4,4,4),
     &        ei22gex(4,4,4,4,4,4,4), ei22uex(4,4,4,4,4,4,4),
     &        e21sex(4,4,4,4,4,4,4),  e21dex(4,4,4,4,4,4,4),
     &        aex, bex, cex, slopeex,
     &        tphbex, hguex, h3cex, haex, hbex, iasyex, ex3
!
        common /stackex/ esex /dangleex/ ed3ex, ed5ex
     &        /HBIex/ ehex, ebex, eiex
     &        /Hstackex/ ehsex /Istackex/ eisex /int11ex/ ei11ex
     &        /int21ex/ e21sex, e21dex/ int22ex
     &        / ei22aex,  ei22cex, ei22gex, ei22uex
     &        /tetraex/ eh4ex / triex /eh3ex
     &        /other/aex, bex, cex, slopeex,
     &        tphbex, hguex, h3cex, haex, hbex, iasyex, ex3
!
        external discrete
!
        binsamp_de=.FALSE.
        binsamp_mbl=.TRUE.
        do_pre_normalize_check=.TRUE.
!
!                set all initials at 0
        seq0(1)=0
        seq0(2)=0
        helixn=0
!
        do i=1, n
          seq1(i,1)=0
          seq1(i,2)=0
          single(i)=0
          ltype(i)=0
          hm(i)=0
          lm(i)=0
        end do
!
        bp(1)=0
        bp(2)=0
!
        do i=1, n6
          probm6(i)=0.d0
        end do
!
        do i=1, n
          probm2(i)=0.d0
          probmbr(i)=0.d0
          probmbl(i)=0.d0
        end do
!
        do i=1, 3*n
          probm1(i)=0.d0
        end do
!
        do i=1, ni
          probmi(i)=0.d0
        end do
!
        do i=1, 3*n
          probmm(i)=0.d0
        end do
!
        probb(1)=0.d0
        probb(2)=0.d0
!
!                multinomial sample w/o knowing if sn.en,
!                with size=2+2*(en-sn-1)+en-sn-2, using s1
        if (ind .eq. 0) then
!
!        IT IS NOT KNOWN IF sn AND en PAIR

!        sumpc=0.d0
!        sumpd=0.d0
!        sumpe=0.d0
!       2010-07-28 (Adam): Modified to match partf.f for n=5
          if (n .eq. 5) then
!          there are only possibilities: open structure or 1 pairs with 5
!          hp 1 pairs with 5 already has a termp (see boundary of partf.f)
             probm1(1)=(u(1,5)-up(1,5))/u(1, 5)
             probm1(2)=up(1, 5)/u(1, 5)
!          all other probabilities stay at 0.
          else
!         == Case a probabilities definition ==
            probm1(1)=pscale(en-sn+1)/u(sn, en)
!         == End of Case a probabilities definition ==
!
!         == Case b probabilities definition ==
            termp=1.d0
            if ( augu(sn, en, n, numseq) ) then
                termp=tphbex
            end if
!
            probm1(2)=(up(sn, en)*termp)/u(sn, en)
!         == End of Case b probabilities definition ==
!
!         == Case d probabilities definition ==
            do k=sn+1, en-1
              termp=1.d0
              if ( augu(sn, k, n, numseq) ) then
                termp=tphbex
              end if
!
              probm1(2+k-sn)=up(sn, k)*termp
!     2010-09-13 (Adam) Removed ed3 here because we don't know if l+1 is avail
     &              *u(k+1, en)/u(sn, en)
!            sumpd=sumpd+probm1(2+k-sn)
            end do
!         == End of Case d probabilities definition ==
!
!         == Case c probabilities definition ==
            do k=sn+1, en-1
              termp=1.d0
              if ( augu(k, en, n, numseq) ) then
                termp=tphbex
              end if
!
              probm1(2+en-sn-1+k-sn)=up(k, en)*termp*pscale(k-sn)
!     2010-09-13 (Adam) Removed ed5 here becuase we don't know if k-1 is avail
!     &          *ed5ex(numseq(k), numseq(en), numseq(k-1))
     &          /u(sn, en)
!            sumpc=sumpc+probm1(2+en-sn-1+k-sn)
            end do
!         == End of Case c probabilities definition ==
!
!         == Case e probabilities definition ==
            do k=sn+1, en-2
              probm1(2+2*(en-sn-1)+k-sn)=pscale(k-sn)*
     &          s1(k,en)/u(sn, en)
!              sumpe=sumpe+probm1(2+2*(en-sn-1)+k-sn)
            end do
!         == End of Case e probabilities definition ==
!
          end if
!
          sump=0.d0
          do i=1, n*3
            sump=sump+probm1(i)
          end do
!
!       2010-07-27 (Adam): sump check should happen before renormalize
          if (do_pre_normalize_check .and.
     &        (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
            write (0, *) "1p, sum=", sump, sn, en
!            write (0,*) "probm1(1)=", probm1(1)
!            write (0,*) "probm1(2)=", probm1(2)
!            write (0,*) "sum probm1(", (2+en-sn-1+(sn+1)-sn) ,
!     &                            "-", (2+en-sn-1+(en-1)-sn) ,")=", sumpc
!            write (0,*) "sum probm1(", (2+(sn+1)-sn) ,
!     &                            "-", (2+(en-1)-sn) ,")=", sumpd
!            write (0,*) "sum probm1(", (2+2*(en-sn-1)+(sn+1)-sn) ,
!     &                            "-", (2+2*(en-sn-1)+(en-2)-sn) ,")=", sumpe
          end if
!
!                renormalize both prob & sump
          do i=1, n*3
            probm1(i)=probm1(i)/sump
          end do
!
          sump=0.d0
          do i=1, n*3
            sump=sump+probm1(i)
          end do
!
          if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
            write (0, *) "1, sum=", sump, sn, en
          end if
!
!            note: probm(i)=0 for i > size of multinomial
          call discrete(3*n, probm1, ir(1))
!
!       ==== Start of case a ===
          if (ir(1) .eq. 1) then
            do i=sn,en
              single(i)=1
              ltype(i)=5
            end do
          end if
!       ===== End of case a ====
!
!       ==== Start of case b ===
          if (ir(1) .eq. 2) then
            seq1(1,1)=sn
            seq1(1,2)=en
          end if
!       ===== End of case b ====
!
!       ==== Start of case d ===
          if (ir(1) .gt. 2 .and. ir(1) .le. (2+en-sn-1)) then
!             2010-06-10 (Adam and Ye): additional binomial sample to decide if l+1
!               is a single-base stack. If so, do the following.
!                1. Set single(l+1), ltype 5
!                2. seq0() lower bound is now l+2, range (l+2,en)
!                    (unless l+2 > en)
!               otherwise, seq0() is (l+1,en), original behavior
            l=sn+ir(1)-2
!           2010-09-02 (Adam) Conditionally perform binomial sampling
            if (binsamp_de) then
              if ((l+4) .le. en) then
                probb(1)=(ed3ex(numseq(sn),numseq(l),numseq(l+1))*
     &                     u(l+2,en)*pscale(1)) /
     &                   (ed3ex(numseq(sn),numseq(l),numseq(l+1))*
     &                     u(l+2,en)*pscale(1)+u(l+1,en)-u(l+2,en)*pscale(1))
!               2010-08-30 (Adam) protect against rounding error
                if (probb(1) .gt. 1.d0) then
                    write(0,*) "Bin prob error in case e: probb(1)=", probb(1)
                    probb(1) = 1.d0
                end if
                if (probb(1) .lt. 0.d0) then
                    write(0,*) "Bin prob error in case e: probb(1)=", probb(1)
                    probb(1) = 0.d0
                end if
                probb(2)=1.d0-probb(1)
              else
!               l+1 can do nothing but single base stack. not enough
!                bases left for anything else.
                probb(1)=1.d0
                probb(2)=0.d0
              end if
!
              call discrete(nb, probb, irbnss1(1))
!
              if (irbnss1(1) .eq. 1) then
!             l+1 is single
                single(l+1)=1
                ltype(l+1)=5
!               Remaining fragment to be handled by main is (l+2, en)
                if ((l+2) .le. en) then
                  seq0(1)=l+2
                  seq0(2)=en
!                 (else, it is a special case where seq0 is empty)
                end if
              else
!               Remaining fragment to be handled by main is (l+1, en)
!               This is the original behavior before binomial sampling implementation.
                if ((l+1) .le. en) then
                  seq0(1)=l+1
                  seq0(2)=en
!                 (else, it is a special case where seq0 is empty)
                end if
              end if
            else
!             Skip binomial sampling (original behavior.)
              if ((l+1) .le. en) then
                seq0(1)=l+1
                seq0(2)=en
!               (else, it is a special case where seq0 is empty)
              end if
            end if
!
!               2010-06-12 (Adam and Ye): This active fragment
!                 is returned regardless of binomial sampling
            seq1(1,1)=sn
            seq1(1,2)=l
!
!             2010-06-10 (Adam and Ye): resetting initial probb vector
            probb(1)=0.d0
            probb(2)=0.d0
!
!            2010-06-10 (Adam and Ye): end of 2010-06-10 mod for case d
!
          end if
!       ===== End of case d ====
!
!       ==== Start of case c ===
          if (ir(1) .gt. (2+en-sn-1) .and. ir(1) .le. (2+2*(en-sn-1))) then
            do i=sn, sn+ir(1)-2-(en-sn-1)-1
                single(i)=1
                ltype(i)=5
            end do
!
            seq1(1,1)=sn+ir(1)-2-(en-sn-1)
            seq1(1,2)=en
!
          end if
!       ===== End of case c ====
!
!       ==== Start of case e ===
          if (ir(1) .gt. (2+2*(en-sn-1))) then
            h=ir(1)+sn-2-2*(en-sn-1)
            do k=h+1, en-1
              termp=1.d0
              if ( augu(h, k, n, numseq) ) then
                termp=tphbex
              end if
!
              probm2(k)=
!     2010-09-13 (Adam) Removed ed5 here becuase we don't know if h-1 is avail
!     2010-09-13 (Adam) Removed ed3 here becuase we don't know if k+1 is avail
     &          up(h, k)*termp*
     &          u(k+1, en)/s1(h, en)
!              sumpc=sumpc+probm1(2+en-sn-1+k-sn)
            end do
!
            sump=0.d0
            do i=1, n
                sump=sump+probm2(i)
            end do
!
!      2010-07-27 (Adam): sump check should happen before renormalize
            if (do_pre_normalize_check .and.
     &          (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
              write (0, *) "1-2p, sum=", sump, sn, en
            end if
!                    renormalize both prb & sump
            do i=1, n
                probm2(i)=probm2(i)/sump
            end do
!
            sump=0.d0
            do i=1, n
                sump=sump+probm2(i)
            end do
!
            if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
             write (0, *) "1-2, sum=", sump, sn, en
            end if
!
!                ir(1) changes value after call for ir2
!                this is bizare!
            call discrete(n, probm2, ir2(1))
!
!           now (h, ir2(1)) is  leftmost BP on (sn,en)
!
            do i=sn, h-1
                single(i)=1
                ltype(i)=5
            end do
!
!             2010-06-10 (Adam and Ye): additional binomial sample to decide if l+1
!               is a single-base stack. If so, do the following.
!                1. Set single(l+1), ltype 5
!                2. seq0() lower bound is now l+2, range (l+2,en)
!                    (unless l+2 > en)
!               otherwise, seq0() is (l+1,en), original behavior
!
            l=ir2(1)
!           2010-09-02 (Adam) Conditionally perform binomial sampling
            if (binsamp_de) then
              if ((l+4) .le. en) then
                probb(1)=(ed3ex(numseq(sn),numseq(l),numseq(l+1))*
     &                     u(l+2,en)*pscale(1)) /
     &                   (ed3ex(numseq(sn),numseq(l),numseq(l+1))*
     &                     u(l+2,en)*pscale(1)+u(l+1,en)-u(l+2,en)*pscale(1))
!               2010-08-30 (Adam) protect against rounding error
                if (probb(1) .gt. 1.d0) then
                    write(0,*) "Bin prob error in case e: probb(1)=", probb(1)
                    probb(1) = 1.d0
                end if
                if (probb(1) .lt. 0.d0) then
                    write(0,*) "Bin prob error in case e: probb(1)=", probb(1)
                    probb(1) = 0.d0
                end if
                probb(2)=1.d0-probb(1)
              else
!               l+1 can do nothing but single base stack. not enough
!                bases left for anything else.
                probb(1)=1.d0
                probb(2)=0.d0
              end if
!
              call discrete(nb, probb, irbnss1(1))
!
              if (irbnss1(1) .eq. 1) then
!             l+1 is single
                single(l+1)=1
                ltype(l+1)=5
!               Remaining fragment to be handled by main is (l+2, en)
                if ((l+2) .le. en) then
                  seq0(1)=l+2
                  seq0(2)=en
!                 (else, it is a special case where seq0 is empty)
                end if
              else
!               Remaining fragment to be handled by main is (l+1, en)
!               This is the original behavior before binomial sampling implementation.
                if ((l+1) .le. en) then
                  seq0(1)=l+1
                  seq0(2)=en
!                 (else, it is a special case where seq0 is empty)
                end if
              end if
            else
!             Skip binomial sampling (original behavior.)
              if ((l+1) .le. en) then
                seq0(1)=l+1
                seq0(2)=en
!               (else, it is a special case where seq0 is empty)
              end if
            end if
!
!               2010-06-12 (Adam and Ye): This active fragment
!                 is returned regardless of binomial sampling
            seq1(1,1)=h
            seq1(1,2)=l
!
!           2010-06-10 (Adam and Ye): resetting initial probb vector
            probb(1)=0.d0
            probb(2)=0.d0
!
!           2010-06-10 (Adam and Ye): end of 2010-06-10 mod for case e
!
          end if
!           ===== End of case e ====
!
          go to 100
!
        end if
!
!    compute uph, ups, upbr, upbl, upi, upm
!         repeat compuation in PF to save memory
!        sn <--> i, en <--> i+inc+1 in partf.f
!
!       IT IS KNOWN THAT sn AND en PAIR
        i=sn
        inc=en-sn-1
!                        uph for hairpin
!       2010-08-17 (Adam) Refactored hpe calculation. Use hpeex() function.
        uph=hpeex(i, i+inc+1, n, numseq, ehex, eh4ex, ehsex,
     &            slopeex, hguex, h3cex, haex, hbex, tphbex)*pscale(inc+2)
!
!                    ups(i, i+inc+1) for stacking
        ups=up(i+1, i+inc)*
     &             esex(numseq(i), numseq(i+1),
     &                numseq(i+inc), numseq(i+inc+1))
        ups=ups*pscale(2)
!
!                        upb(i, i+inc+1) for bulges
!                l starts @ i+5, for >= 3nt for hairpin (i+1)--l
        upbr=0.d0
        do l=i+5, i+inc-1
!                add stacking if size=1, to loop penalty
!                or au,gu penalty if size >= 2
          prod1=1.d0
          prod2=1.d0
          if ( augu(i, i+inc+1, n, numseq) ) then
            prod1=tphbex
          end if
          if ( augu(l, i+1, n, numseq) ) then
            prod2=tphbex
          end if
!
          if ((i+inc-l) .gt. 30) then
            upbr=upbr+prod1*prod2*
     &             ebex(30)*( slopeex**dlog(dble(i+inc-l)/30.d0) )
!     &         ebex(30)*dexp( -(1000.d0/(1.9872d0*310.15d0))
!     &            *1.079d0*dlog(dble(i+inc-l)/30.d0) )
     &            *up(i+1,l)*pscale(i+inc+2-l)
!
          else
            if (l .eq. (i+inc-1)) then
                upbr=upbr+ebex(1)*up(i+1,l)*
     &             esex(numseq(i), numseq(i+1),
     &                numseq(i+inc-1), numseq(i+inc+1))
     &            *pscale(i+inc+2-l)
            else
                upbr=upbr+prod1*prod2*ebex(i+inc-l)
     &                *up(i+1,l)
     &            *pscale(i+inc+2-l)
            end if
          end if
        end do
!
        upbl=0.d0
!              h ends @ i+inc-4, for >= 3nt for h--(i+inc) hairpin
        do h=i+2,  i+inc-4
!         add stacking if size=1, or au,gu penalty if size >= 2
          prod1=1.d0
          prod2=1.d0
          if ( augu(i, i+inc+1, n, numseq) ) then
            prod1=tphbex
          end if
          if  ( augu(i+inc, h, n, numseq) ) then
            prod2=tphbex
          end if
!
          if ((h-i-1) .gt. 30) then
            upbl=upbl+prod1*prod2*
     &        ebex(30)*( slopeex**dlog(dble(h-i-1)/30.d0) )
!     &          ebex(30)*dexp( -(1000.d0/(1.9872d0*310.15d0))
!     &            *1.079d0*dlog(dble(i+inc-l)/30.d0) )
     &      *up(h,i+inc)*pscale(h-i+1)
          else
            if (h .eq. (i+2)) then
                upbl=upbl+ebex(h-i-1)*up(h,i+inc)*
     &                       esex(numseq(i), numseq(i+2),
     &                numseq(i+inc), numseq(i+inc+1))
     &            *pscale(h-i+1)
            else
                upbl=upbl+prod1*prod2*ebex(h-i-1)
     &            *up(h,i+inc)
     &            *pscale(h-i+1)
            end if
          end if
        end do
!
!                              upi(i, i+inc+1) for interior
        upi=0.d0
        do h=i+2, min(i+30, i+inc-5)
          do l=max(h+4, h+inc-31), i+inc-1
!
!                    1x1 I loop, use 1x1 table
            if (h .eq. (i+2) .and. l .eq. (i+inc-1)) then
              upi=upi+ei11ex(numseq(i), numseq(i+inc+1),
     &                numseq(i+2), numseq(i+inc-1),
     &                numseq(i+1), numseq(i+inc))*up(h,l)
     &            *pscale(h+inc+1-l)
              go to 6
            end if
!
!                1x2 loop, single nt on 5' --> 3'
            if (h .eq. (i+2) .and. l .eq. (i+inc-2)) then
              upi=upi+
     &            e21sex(numseq(i), numseq(i+inc+1),
     &            numseq(i+2), numseq(i+inc-2),
     &            numseq(i+1),
     &            numseq(i+inc-1), numseq(i+inc))*up(h,l)
     &            *pscale(h+inc+1-l)
              go to 6
            end if
!
!                2x1 loop, double nt on 5' --> 3'
            if (h .eq. (i+3) .and. l .eq. (i+inc-1)) then
              upi=upi+
     &            e21dex(numseq(i), numseq(i+inc+1),
     &            numseq(i+3), numseq(i+inc-1),
     &            numseq(i+1), numseq(i+2),
     &            numseq(i+inc))*up(h,l)
     &            *pscale(h+inc+1-l)
              go to 6
            end if
!
!                    2x2 loop
            if (h .eq. (i+3) .and. l .eq. (i+inc-2)) then
              if (numseq(i+inc) .eq. 1) then
                upi=upi+
     &            ei22aex(numseq(i), numseq(i+inc+1),
     &            numseq(i+3), numseq(i+inc-2),
     &            numseq(i+1), numseq(i+2),
     &            numseq(i+inc-1))*up(h,l)
     &            *pscale(h+inc+1-l)
              end if
!
              if (numseq(i+inc) .eq. 2) then
                upi=upi+
     &            ei22cex(numseq(i), numseq(i+inc+1),
     &            numseq(i+3), numseq(i+inc-2),
     &            numseq(i+1), numseq(i+2),
     &            numseq(i+inc-1))*up(h,l)
     &            *pscale(h+inc+1-l)
              end if
!
              if (numseq(i+inc) .eq. 3) then
                upi=upi+
     &            ei22gex(numseq(i), numseq(i+inc+1),
     &            numseq(i+3), numseq(i+inc-2),
     &            numseq(i+1), numseq(i+2),
     &            numseq(i+inc-1))*up(h,l)
     &            *pscale(h+inc+1-l)
              end if
!
              if (numseq(i+inc) .eq. 4) then
                upi=upi+
     &            ei22uex(numseq(i), numseq(i+inc+1),
     &            numseq(i+3), numseq(i+inc-2),
     &            numseq(i+1), numseq(i+2),
     &            numseq(i+inc-1))*up(h,l)
     &            *pscale(h+inc+1-l)
              end if
!
              go to 6
            end if
!
!                all other cases, > 4 nt and 1x3, 3x1 (4 nt)
            if ( ((h+inc-l-1) .gt. 4 ) .or.
     &                  (h .eq. (i+2) .and. l .eq. (i+inc-3)) .or.
     &             ( h .eq. (i+4) .and. l .eq. (i+inc-1))  ) then
!
!                2 terminal stacks +asym. penalty,maximum is 3
!        this limit is used by mfold, but not made clear by DHM
!                   for 1Xn,nX1 treat 1 nt as 2 same nt(DHM email 1/7/99)
!               and treat the 3 nts involved in 2 mismatches as A
!               (DHM email 2/8/99) i.e, take numseq(*)=1
!
              if (h .eq. (i+2) .or. l .eq. (i+inc-1)) then
                prod1=eisex(numseq(i), 1,
     &                                1, numseq(i+inc+1))
     &            *dmax1(ex3, iasyex**abs(dble((h-i-1)-(i+inc-l))) )
!        dmax1 gives same result as min for the original=cc
!     &      *dexp(-(1000.d0/(1.9872d0*310.15d0))
!     &        *min(3.d0, 0.5d0*abs(dble((h-i-1)-(i+inc-l)))))
!!   &      *dexp(-scale*min(3.d0, iasy*abs(dble((h-i-1)-(i+inc-l)))))
     &            *eisex(numseq(l), 1, 1, numseq(h))
!
              else
!
                prod1=eisex(numseq(i),
     &          numseq(i+1), numseq(i+inc), numseq(i+inc+1))
     &            *dmax1(ex3, iasyex**abs(dble((h-i-1)-(i+inc-l))) )
!        dmax1 gives same result as min forthe original=cc
!     &      *dexp(-(1000.d0/(1.9872d0*310.15d0))
!     &        *min(3.d0, 0.5d0*abs(dble((h-i-1)-(i+inc-l)))))
!!   &    *dexp(-scale*min(3.d0, iasy*abs(dble((h-i-1)-(i+inc-l)))))
     &            *eisex(numseq(l), numseq(l+1),
     &                numseq(h-1), numseq(h))
              end if
              upi=upi+eiex(h+inc-l-1)*up(h,l)
     &               *prod1
     &            *pscale(h+inc+1-l)
!
            end if
!
6           continue
!
          end do
        end do
!
!                    compute upm
        sum1=0.d0
        do l=i+2, i+inc
!                 add au/gu terminal penalty
          termp=1.d0
          if ( augu(i+1, l, n, numseq) )  then
            termp=tphbex
          end if
!
          sum1=sum1+up(i+1,l)*termp*
     &         (  ed3ex(numseq(i+1),numseq(l),numseq(l+1))
     &                *u1(l+2, i+inc)*pscale(1)
     &           +u1(l+1, i+inc)-u1(l+2, i+inc)*pscale(1)  )
        end do
        sum1=sum1*pscale(2)
!
        sum2=0.d0
        do l=i+3, i+inc
!                 add au/gu terminal penalty
          termp=1.d0
          if ( augu(i+2, l, n, numseq) ) then
            termp=tphbex
          end if
!
!                5's single stack is ignored but included in c.
          sum2=sum2+
!     &  dmax1(ed3ex(numseq(i+inc+1), numseq(i), numseq(i+1),
!     &            ed5ex(numseq(i+2),numseq(l),numseq(i+1))))
     &      ed3ex(numseq(i+inc+1), numseq(i), numseq(i+1))
     &         *bex*up(i+2,l)*termp*
     &        ( ed3ex(numseq(i+2),numseq(l),numseq(l+1))
     &            *u1(l+2, i+inc)*pscale(1)
     &           +u1(l+1, i+inc)-u1(l+2, i+inc)*pscale(1)  )
!
        end do
        sum2=sum2*pscale(3)
!
        sum3=0.d0
        do h=i+3, i+inc-1
          sum3=sum3+pscale(h-i)*
     &        (bex**dble(h-i-1))*s2(h, i+inc+1)
        end do
        sum3=sum3*ed3ex(numseq(i+inc+1),numseq(i),numseq(i+1))
!
!     2010-09-24 (Adam&Ye) termp on the upm term as well (not in paper A2)
        termp=1.d0
        if ( augu(i,i+inc+1,n,numseq) )  then
          termp=tphbex
        end if
!       2010-10-20 (Adam) Factored out common terms in large sum.
        upm=aex*(cex**2.d0)*(sum1+sum2+sum3)*termp
!
!         end of computing uph, ups, upbr, upbl, upi, upm sample given base
!         pair  sn.en, first
!
!                compute probm6(6) for
!                1-h, 2-stack, 3-br, 4-bl, 5-I and 6-M
        probm6(1)=uph/up(sn,en)
        probm6(2)=ups/up(sn,en)
        probm6(3)=upbr/up(sn,en)
        probm6(4)=upbl/up(sn,en)
        probm6(5)=upi/up(sn,en)
        probm6(6)=upm/up(sn,en)
!
        sump=probm6(1)+probm6(2)+probm6(3)+
     &        probm6(4) +probm6(5)+probm6(6)
!              write (0, *) sump
        if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
          write (0, *) "Up(",sn,",",en,")=",up(sn,en)
          write (0, *) "(uph,ups,upbr,upbl,upi,upm)=",
     &                 uph,ups,upbr,upbl,upi,upm
          write (0, *) "probm6(1-6)=",
     &                 probm6(1),probm6(2),probm6(3),probm6(4),probm6(5),probm6(6)
          write (0, *) "2, sum=", sump, sn, en
        end if
!
!        now sample one of H, Stack, Br, Bl, I or M
!
        call discrete(n6, probm6,  ir6(1))
!
!                 case 1:    H-loop is sampled
        if (ir6(1) .eq. 1) then
          bp(1)=sn
          bp(2)=en
!
          do i=sn+1, en-1
            single(i)=1
            ltype(i)=1
          end do
        end if
!
!                case 2: stacking is sampled
        if (ir6(1) .eq. 2) then
          bp(1)=sn
          bp(2)=en
          seq1(1,1)=sn+1
          seq1(1,2)=en-1
        end if
!
!                case 3: B-loop    with right single bases
        if (ir6(1) .eq. 3) then
!
          do l=sn+5, en-2
!                add stacking if size=1, to loop penalty
!                or au,gu penalty if size >= 2
            prod1=1.d0
            prod2=1.d0
            if ( augu(sn, en, n, numseq) ) then
                prod1=tphbex
            end if
            if ( augu(l, sn+1, n, numseq) ) then
                prod2=tphbex
            end if
!
            if ((en-1-l) .gt. 30) then
                probmbr(l)=prod1*prod2*pscale(1+en-l)*
     &            ebex(30)*(slopeex**dlog(dble(en-1-l)/30.d0))*
     &            up(sn+1,l)/upbr
            else
              if (l .eq. (en-2)) then
                probmbr(l)=ebex(1)*up(sn+1,l)*
     &             esex(numseq(sn), numseq(sn+1),
     &              numseq(en-2), numseq(en))
     &            *pscale(3)/upbr
              else
                        probmbr(l)=prod1*prod2*ebex(en-1-l)*
     &            up(sn+1,l)*pscale(1+en-l)/upbr
              end if
            end if
          end do
!
          sump=0.d0
          do i=1, n
            sump=sump+probmbr(i)
          end do
!
          if (do_pre_normalize_check .and.
     &        (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
            write (0, *) "3p, sum=", sump, sn, en, upbr
          end if
!
!            renormalize both prob. and sump
          do i=1, n
            probmbr(i)=probmbr(i)/sump
          end do
!
          sump=0.d0
          do i=1, n
            sump=sump+probmbr(i)
          end do
!
          if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
            write (0, *) "3, sum=", sump, sn, en, upbr
          end if
!
          call discrete(n, probmbr, irbr(1))
!
!                              irbr(1) is sampled l
!                            output
          bp(1)=sn
          bp(2)=en
!
          do i=irbr(1)+1, en-1
            single(i)=1
            ltype(i)=2
          end do
!
          seq1(1,1)=sn+1
          seq1(1,2)=irbr(1)
!
        end if
!
!                case 4: B-loop    with left single bases
        if (ir6(1) .eq. 4) then
          do h=sn+2,  en-5
!            add stacking if size=1, or au,gu penalty if size >= 2
            prod1=1.d0
            prod2=1.d0
            if  ( augu(sn, en, n, numseq) )  then
                prod1=tphbex
            end if
!
            if  ( augu(en-1, h, n, numseq) )  then
                prod2=tphbex
            end if
!
            if ((h-sn-1) .gt. 30) then
                probmbl(h)=prod1*prod2*pscale(h-sn+1)*
     &            ebex(30)*(slopeex**dlog(dble(h-sn-1)/30.d0))*
     &            up(h,en-1)/upbl
            else
              if (h .eq. (sn+2)) then
                probmbl(h)=ebex(h-sn-1)*up(h,en-1)*
     &             esex(numseq(sn), numseq(sn+2),
     &            numseq(en-1), numseq(en))
     &            *pscale(3)/upbl
              else
                probmbl(h)=prod1*prod2*ebex(h-sn-1)*
     &            up(h,en-1)*pscale(h-sn+1)/upbl
              end if
            end if
          end do
!
          sump=0.d0
          do i=1, n
            sump=sump+probmbl(i)
          end do
!
          if (do_pre_normalize_check .and.
     &        (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
            write (0, *) "4p, sum=", sump, sn, en
          end if
!
!                renormalize both prb & sump
          do i=1, n
            probmbl(i)=probmbl(i)/sump
          end do
!
          sump=0.d0
          do i=1, n
            sump=sump+probmbl(i)
          end do
!
          if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
            write (0, *) "4, sum=", sump, sn, en
          end if
!
          call discrete(n, probmbl, irbl(1))
!
!                            output
          bp(1)=sn
          bp(2)=en
!
          do i=sn+1,irbl(1)-1
            single(i)=1
            ltype(i)=2
          end do
!
          seq1(1,1)=irbl(1)
          seq1(1,2)=en-1
!
        end if
!
!                case 5: I-loop is sampled
        if (ir6(1) .eq. 5) then
!    write(0, *) "I-loop!", "sn=", sn, "en=", en, "ind=", ind
!    write(0, *) "P(stack)=", probm6(2)
!    write(0, *)  "P(B-rss)", probm6(3)
!    write(0, *)  "P(B-lss)", probm6(4)
!    write(0, *)  "P(I-loop)", probm6(5)
!
          do h=sn+2, min(sn+30, en-6)
            do l=max(h+4, h-sn-30+en-2), en-2
              k=0
              do h0=sn+2, h-1
                k=k+en-2-max(h0+4, h0-sn-30+en-2)+1
              end do
              k=k+l-max(h+4, h-sn-30+en-2)+1
!
!                    1x1 I loop, use 1x1 table
              if (h .eq. (sn+2) .and. l .eq. (en-2)) then
                probmi(k)=
     &          ei11ex(numseq(sn), numseq(en),
     &                numseq(sn+2), numseq(en-2),
     &                numseq(sn+1), numseq(en-1) )*
     &              up(h,l)*pscale(h-sn+en-l)/upi
              end if
!
!                2x1 loop, single nt on 5' --> 3'
              if (h .eq. (sn+2) .and. l .eq. (en-3)) then
                probmi(k)=
     &            e21sex(numseq(sn), numseq(en),
     &            numseq(sn+2), numseq(en-3),
     &            numseq(sn+1),
     &            numseq(en-2), numseq(en-1))*
     &            up(h,l)*pscale(h-sn+en-l)/upi
              end if
!
!                2x1 loop, double nt on 5' --> 3'
              if (h .eq. (sn+3) .and. l .eq. (en-2)) then
                probmi(k)=
     &            e21dex(numseq(sn), numseq(en),
     &            numseq(sn+3), numseq(en-2),
     &            numseq(sn+1), numseq(sn+2),
     &            numseq(en-1))*up(h,l)*
     &            pscale(h-sn+en-l)/upi
              end if
!
!                    2x2 loop
              if (h .eq. (sn+3) .and. l .eq. (en-3)) then
                if (numseq(en-1) .eq. 1) then
                  probmi(k)=
     &              ei22aex(numseq(sn), numseq(en),
     &              numseq(sn+3), numseq(en-3),
     &              numseq(sn+1), numseq(sn+2),
     &              numseq(en-2))
     &              *up(h,l)*pscale(h-sn+en-l)/upi
                end if
!
                if (numseq(en-1) .eq. 2) then
                  probmi(k)=
     &              ei22cex(numseq(sn), numseq(en),
     &              numseq(sn+3), numseq(en-3),
     &              numseq(sn+1), numseq(sn+2),
     &              numseq(en-2))
     &              *up(h,l)*pscale(h-sn+en-l)/upi
                end if
!
                if (numseq(en-1) .eq. 3) then
                  probmi(k)=
     &              ei22gex(numseq(sn), numseq(en),
     &              numseq(sn+3), numseq(en-3),
     &              numseq(sn+1), numseq(sn+2),
     &              numseq(en-2))
     &              *up(h,l)*pscale(h-sn+en-l)/upi
                end if
!
                if (numseq(en-1) .eq. 4) then
                  probmi(k)=
     &              ei22uex(numseq(sn), numseq(en),
     &              numseq(sn+3), numseq(en-3),
     &              numseq(sn+1), numseq(sn+2),
     &              numseq(en-2))
     &              *up(h,l)*pscale(h-sn+en-l)/upi
                end if
!
              end if
!
!                all other cases, > 4 nt and 1x3, 3x1 (4 nt)
              if ( ((h+en-sn-1-l-1) .gt. 4 ) .or.
     &                  (h .eq. (sn+2) .and. l .eq. (en-4)) .or.
     &             ( h .eq. (sn+4) .and. l .eq. (en-2))  ) then
!
!                2 terminal stacks +asym. penalty (maximum 3)
!                 for 1Xn,nX1 treat 1 nt as 2 same nt(DHM email 1/7/99)
!               and treat the 3 nts involved in 2 mismatches as A
!               (DHM email 2/8/99) i.e, take numseq(*)=1
!
                if (h .eq. (sn+2) .or. l .eq. (en-2)) then
                        prod1=eisex(numseq(sn), 1, 1, numseq(en))
     &  *dmax1( ex3, iasyex**abs(dble((h-sn-1)-(en-1-l)))  )
!     &    *dexp(-scale*min(3.d0, iasy*abs(dble((h-sn-1)-(en-1-l)))))
     &    *eisex(numseq(l), 1, 1, numseq(h))
!
                else
!
                  prod1=eisex(numseq(sn),
     &             numseq(sn+1),numseq(en-1), numseq(en))
     &  *dmax1( ex3, iasyex**abs(dble((h-sn-1)-(en-1-l)))  )
!     &      *dexp(-scale*min(3.d0, iasy*abs(dble((h-sn-1)-(en-1-l)))))
     &      *eisex(numseq(l), numseq(l+1),
     &                    numseq(h-1), numseq(h))
                end if
!
                    probmi(k)=eiex(h+en-sn-1-l-1)*up(h,l)
     &               *prod1*pscale(h-sn+en-l)/upi
              end if
!
            end do
          end do
!
          sump=0.d0
          do i=1, ni
            sump=sump+probmi(i)
          end do
!
          if (do_pre_normalize_check .and.
     &        (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
            write (0, *) "5p, sum=", sump, sn, en
          end if
!
!                renormalize both prb & sump
          do i=1, ni
            probmi(i)=probmi(i)/sump
          end do
!
          sump=0.d0
          do i=1, ni
            sump=sump+probmi(i)
          end do
!
          if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
            write (0, *) "5, sum=", sump, sn, en
          end if
!
          call discrete(ni, probmi,  iri(1))
!
!                    now find h, l with index iri(1)
          do h=sn+2, min(sn+30, en-6)
            do l=max(h+4, h-sn-30+en-2), en-2
              k=0
              do h0=sn+2, h-1
                k=k+en-2-max(h0+4, h0-sn-30+en-2)+1
              end do
              k=k+l-max(h+4, h-sn-30+en-2)+1
              if (k .eq. iri(1)) then
                go to 1
              end if
            end do
          end do
1         continue
!
!                            output
          bp(1)=sn
          bp(2)=en
!
          do i=sn+1, h-1
            single(i)=1
            ltype(i)=3
          end do
!
          do i=l+1, en-1
            single(i)=1
            ltype(i)=3
          end do
!
          seq1(1,1)=h
          seq1(1,2)=l
!
        end if
!
!                case 6: M-loop sampled
        if (ir6(1) .eq. 6) then
!         2010-10-20 (Adam) Cancel the effects of termp(sn,en), aex, and
!           (cex**2.d0), which are all constant for all terms in upm(sn,en)
          termp=1.d0
          if ( augu(sn, en, n, numseq) )  then
            termp=tphbex
          end if
          upm = upm / termp / aex / (cex**2.d0)
!
          do l=sn+2, en-1
!                 add au/gu terminal penalty
            termp=1.d0
            if ( augu(sn+1, l, n, numseq) )  then
                termp=tphbex
            end if
!
            probmm(l)=pscale(2)*
     &        up(sn+1,l)*termp*
     &          ( ed3ex(numseq(sn+1),numseq(l),numseq(l+1))
     &                *u1(l+2, en-1)*pscale(1)
     &           +u1(l+1, en-1)-u1(l+2, en-1)*pscale(1) )
     &        /upm
!
          end do
!
          do l=sn+3, en-1
!                 add au/gu terminal penalty
            termp=1.d0
            if ( augu(sn+2, l, n, numseq) )  then
                termp=tphbex
            end if
!
!            5's single stack is ignored but included in c.
            probmm(n+l)=pscale(3)*
!     &      dmax1(ed3ex(numseq(en), numseq(sn), numseq(sn+1),
!     &            ed5ex(numseq(sn+2),numseq(l),numseq(sn+1)))
     &      ed3ex(numseq(en), numseq(sn), numseq(sn+1))
     &         *bex*up(sn+2,l)*termp*
     &         ( ed3ex(numseq(sn+2),numseq(l),numseq(l+1))
     &            *u1(l+2, en-1)*pscale(1)
     &           +u1(l+1, en-1)-u1(l+2,en-1)*pscale(1)  )
     &        /upm
!
          end do
!
          do h=sn+3, en-2
            probmm(2*n+h)=pscale(h-sn)*
     &         (bex**dble(h-sn-1))*s2(h, en)
     &        *ed3ex(numseq(en),numseq(sn),numseq(sn+1))
     &        /upm
          end do
!
          sump=0.d0
          do i=1, 3*n
            sump=sump+probmm(i)
          end do
!
          if (do_pre_normalize_check .and.
     &        (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
            write (0, *) "6p, sum=", sump, sn, en
          end if
!
!                renormalize prb & sump
          do i=1, 3*n
            probmm(i)=probmm(i)/sump
          end do
!
          sump=0.d0
          do i=1, 3*n
            sump=sump+probmm(i)
          end do
!
          if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
            write (0, *) "6, sum=", sump, sn, en, upm
          end if
!
          call discrete(3*n, probmm, irm(1))
!
!                    reset initials for reuse
          do i=1, 3*n
            probmm(i)=0.d0
          end do
!
          if (irm(1) .ge. (sn+2) .and. irm(1) .le. (en-1) ) then
            h=sn+1
            l=irm(1)
          end if
!
          if (irm(1) .ge. (n+sn+3) .and. irm(1) .le. (n+en-1) ) then
            h=sn+2
            l=irm(1)-n
          end if
!
          if (irm(1) .ge. (2*n+sn+3) .and. irm(1) .le. (2*n+en-2) ) then
            h=irm(1)-2*n
            do i=1, n
                probm2(i)=0.d0
            end do
!
            do l=h+1, en-1
              termp=1.d0
              if ( augu(h, l, n, numseq) ) then
                termp=tphbex
              end if
              probm2(l)=pscale(1)*
     &          up(h,l)*termp*
     &          ed5ex(numseq(h),numseq(l),numseq(h-1))*
     &          ( ed3ex(numseq(h),numseq(l),numseq(l+1))
     &            *u1(l+2, en-1)*pscale(1)
     &          +u1(l+1, en-1)-u1(l+2, en-1)*pscale(1)  )
     &          /s2(h, en)
            end do
!
            sump=0.d0
            do i=1, n
                sump=sump+probm2(i)
            end do
!
            if (do_pre_normalize_check .and.
     &          (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
              write (0, *) "6-2p, sum=", sump, sn, en
            end if
!
!                    renormalize prb & sump
            do i=1, n
                probm2(i)=probm2(i)/sump
            end do
!
            sump=0.d0
            do i=1, n
                sump=sump+probm2(i)
            end do
!
            if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
                write (0, *) "6-2, sum=", sump, sn, en
            end if
!
            call discrete(n, probm2,  ir2(1))
!
            l=ir2(1)
!
          end if
!
          helixn=1
          hm(helixn)=h
          lm(helixn)=l
!                        output 1
          bp(1)=sn
          bp(2)=en
!                        (sn+1, h-1) unpaired
          if (h .ge. (sn+2)) then
            do i=sn+1, h-1
                single(i)=1
                ltype(i)=4
            end do
          end if
!
          seq1(1,1)=h
          seq1(1,2)=l
!
!       write (0, *) h, l
!                sample 2nd (or more) helx with U1(i,j)
!                5'->3', opposite to sample98.f & McCaskill
!
          i0=l+1
          j0=en-1
!
5         continue
!
!         2010-10-19 (Adam) Perform binomial sampling to decide if i0 dangles.
          if ((i0+4) .le. j0) then
            probb(1)=(ed3ex(numseq(hm(helixn)),numseq(lm(helixn)),numseq(i0))*
     &                u1(i0+1,j0)*pscale(1)) /
     &               (ed3ex(numseq(hm(helixn)),numseq(lm(helixn)),numseq(i0))*
     &                 u1(i0+1,j0)*pscale(1) + u1(i0,j0) - u1(i0+1,j0)*pscale(1))
            probb(2)=1.d0-probb(1)
          else
!           i0 can do nothing but single base stack. not enough
!           bases left for anything else.
            probb(1)=1.d0
            probb(2)=0.d0
          end if
!
          call discrete(nb, probb, irbnss1(1))
!
          if (irbnss1(1) .eq. 1) then
!           i0 3' dangles on hm(helixn), lm(helixn)
!           Set i0 unpaired, increment i0, and sample as before.
            single(i0)=1
            ltype(i0)=4
            i0=i0+1
!
            do l=i0+1, j0
              stack3=1.d0
              scaleterm=1.d0
              if (l .ne. j0) then
!               assert (l .le. (j0-1))
                stack3=ed3ex(numseq(i0), numseq(l),numseq(l+1))
                scaleterm=pscale(j0-l)
              end if
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i0, l, n, numseq) ) then
                  termp=tphbex
              end if
!
              if ((l+1) .gt. n) then
                probmm(l)=cex*up(i0,l)*termp
     &                   *stack3*(bex**dble(j0-l))*scaleterm
     &                   / u1(i0,j0)
              else
                probmm(l)=cex*up(i0,l)*termp
     &                   *(stack3*(bex**dble(j0-l))*scaleterm
     &                     +ed3ex(numseq(i0),numseq(l),numseq(l+1))
     &                     *u1(l+2, j0)*pscale(1)
     &                     +u1(l+1, j0) - u1(l+2, j0)*pscale(1))
     &                   / u1(i0,j0)
              end if
            end do
!
            do l=i0+2, j0
              stack3=1.d0
              scaleterm=1.d0
              if (l .ne. j0) then
!               assert (l .le. (j0-1))
                stack3=ed3ex(numseq(i0+1),numseq(l),numseq(l+1))
                scaleterm=pscale(j0-l)
              end if
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i0+1, l, n, numseq) )  then
                  termp=tphbex
              end if
!
              if ((l+1) .gt. n) then
                probmm(n+l)=pscale(1)*cex*bex*up(i0+1,l)*termp
     &                     *(stack3*(bex**dble(j0-l)))*scaleterm
     &                     / u1(i0,j0)
              else
                probmm(n+l)=pscale(1)*cex*bex*up(i0+1,l)*termp
     &                     *(stack3*(bex**dble(j0-l))*scaleterm
     &                       +ed3ex(numseq(i0+1),numseq(l),numseq(l+1))
     &                       *u1(l+2, j0)*pscale(1)
     &                       +u1(l+1, j0)-u1(l+2, j0)*pscale(1))
     &                     / u1(i0,j0)
              end if
            end do
!
            do h=i0+2, j0-1
              probmm(2*n+h)=pscale(h-i0)*cex*(bex**dble(h-i0))*s3(h, j0)
     &                     / u1(i0,j0)
            end do
!
            sump=0.d0
            do i=1, 3*n
              sump=sump+probmm(i)
            end do
!
            if (do_pre_normalize_check .and.
     &          (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
              write (0, *) "7ap, sum=", sump, sn, en
            end if
!                    renormalize prb & sump
            do i=1, 3*n
              probmm(i)=probmm(i)/sump
            end do
!
            sump=0.d0
            do i=1, 3*n
              sump=sump+probmm(i)
            end do
!
            if (sump .gt. 1.000000001d0 .or. sump .lt. 0.999999d0) then
              write (0, *) "7a, sum=", sump, i0, j0
            end if
!
            call discrete(3*n, probmm,  irm(1))
!
            do i=1, 3*n
              probmm(i)=0.d0
            end do
!
            if (irm(1) .ge. (i0+1) .and. irm(1) .le. j0 ) then
              h=i0
              l=irm(1)
            else if (irm(1) .ge. (n+i0+2) .and. irm(1) .le. (n+j0) ) then
              h=i0+1
              l=irm(1)-n
            else
!             assert (irm(1) .ge. (2*n+i0+2) .and. irm(1) .le. (2*n+j0-1) )
              h=irm(1)-2*n
              do i=1, n
                  probm2(i)=0.d0
              end do
!
              do l=h+4, j0
                stack3=1.d0
                scaleterm=1.d0
                if (l .ne. j0) then
!                 assert (l .le. (j0-1))
                  stack3=ed3ex(numseq(h), numseq(l),numseq(l+1))
                  scaleterm=pscale(j0-l)
                end if
!                 add au/gu terminal penalty
                termp=1.d0
                if ( augu(h, l, n, numseq) ) then
                  termp=tphbex
                end if
!
                if ((l+1) .gt. n) then
                  probm2(l)=up(h, l)*termp
     &                      *ed5ex(numseq(h), numseq(l), numseq(h-1))
     &                      *(stack3*(bex**dble(j0-l)))*scaleterm
     &                      / s3(h,j0)
                else
                  probm2(l)=up(h, l)*termp
     &                     *ed5ex(numseq(h), numseq(l), numseq(h-1))
     &                     *(stack3*(bex**dble(j0-l))*scaleterm
     &                       +ed3ex(numseq(h),numseq(l),numseq(l+1))
     &                       *u1(l+2, j0)*pscale(1)
     &                       +u1(l+1, j0)-u1(l+2, j0)*pscale(1) )
     &                     / s3(h,j0)
                end if
              end do
!
              sump=0.d0
              do i=1, n
                sump=sump+probm2(i)
              end do
!
              if (do_pre_normalize_check .and.
     &            (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0)) then
                write (0, *) "7a-2p, sum=", sump, sn, en
              end if

!                renormalize prb & sum
              do i=1, n
                probm2(i)=probm2(i)/sump
              end do
!
              sump=0.d0
              do i=1, n
                sump=sump+probm2(i)
              end do
!
              if (sump .gt. 1.0000001d0 .or. sump .lt. 0.999999d0) then
                write (0, *) "7a-2, sum=", sump, sn, en
              end if
!
              call discrete(n, probm2, ir2(1))
!
              l=ir2(1)
            end if
          else
!           i0 is paired.
            do l=i0+1, j0
              stack3=1.d0
              scaleterm=1.d0
              if (l .ne. j0) then
!               assert (l .le. (j0-1))
                stack3=ed3ex(numseq(i0), numseq(l),numseq(l+1))
                scaleterm=pscale(j0-l)
              end if
!                 add au/gu terminal penalty
              termp=1.d0
              if ( augu(i0, l, n, numseq) ) then
                  termp=tphbex
              end if
!
              if ((l+1) .gt. n) then
                probmm(l)=cex*up(i0,l)*termp
     &                   *stack3*(bex**dble(j0-l))*scaleterm
     &                   / u1(i0,j0)
              else
                probmm(l)=cex*up(i0,l)*termp
     &                   *(stack3*(bex**dble(j0-l))*scaleterm
     &                     +ed3ex(numseq(i0),numseq(l),numseq(l+1))
     &                     *u1(l+2, j0)*pscale(1)
     &                     +u1(l+1, j0) - u1(l+2, j0)*pscale(1))
     &                   / u1(i0,j0)
              end if
            end do
!
            sump=0.d0
            do i=1, 3*n
              sump=sump+probmm(i)
            end do
!           2010-10-21 (Adam) pre_normalize_check is not possible, here,
!            because we are dealing with only part of the sum (sum1 only).
!            Instead, just make sure that sump is not 0. This would mean
!            that there is no possible candidate for i0's partner,
!            even though it was somehow sampled as paired. Even though
!            it indicates a logic error, to do otherwise leads to segfault.
            if (sump .eq. 0.d0) then
!             Ooops. There's no possible partner for i0. How was it that
!             i0 was determined to be paired through binary sampling?
              write (0,*) "i0 cannot pair with anything from ",(i0+1), "to", j0, "."
              do i=i0+1, j0
                single(i)=1
                ltype(i)=4
              end do
              go to 100
            end if
!
            do i=1, 3*n
              probmm(i)=probmm(i)/sump
            end do
!
            sump=0.d0
            do i=1, 3*n
              sump=sump+probmm(i)
            end do
!
            if (sump .gt. 1.000000001d0 .or. sump .lt. 0.999999d0) then
              write (0, *) "7b, sum=", sump, i0, j0
            end if
!
            call discrete(3*n, probmm,  irm(1))
!
            do i=1, 3*n
              probmm(i)=0.d0
            end do
!
            if (irm(1) .ge. (i0+1) .and. irm(1) .le. j0 ) then
              h=i0
              l=irm(1)
            else
              write (0, *) "Error 7b. Sampled outside of range."
            end if
          end if
!
!                        more output
          helixn=helixn+1
          hm(helixn)=h
          lm(helixn)=l
          if (h .ge. (i0+1)) then
            do i=i0, h-1
                single(i)=1
                ltype(i)=4
            end do
          end if
!
          seq1(helixn, 1)=hm(helixn)
          seq1(helixn, 2)=lm(helixn)
!                    sample binomial for 0 or more helix
!                note now h=hm(helixn), l=lm(helixn) in stack3!
!          if (l .eq. j0) then
          stack3=1.d0
!          end if
!
!            in commented out parts of next 2 ifs, ed5(*,-) for 5'
!            stack on 3' end of closing BP of M-loop,
!            * to be simulated or average used, it is ignored now
!
          if (l .eq. (j0-1)) then
            stack3=ed3ex(numseq(h), numseq(l),numseq(l+1))
!      stack3=dmax1(ed3ex(numseq(h),numseq(l),numseq(l+1)),
!     &                 ed5ex(numseq(*),numseq(j0+1), numseq(j0))
          end if
!
          if (l .le. (j0-2)) then
            stack3=ed3ex(numseq(h),numseq(l),numseq(l+1))
!      &    ed3ex(numseq(h),numseq(l),numseq(l+1))
!      &        *ed5ex(numseq(*),numseq(j0+1), numseq(j0))
          end if
!
          if (lm(helixn) .eq. j0) then
            scaleterm=1.d0
          else
            scaleterm=pscale(j0-lm(helixn))
          end if
!
          probb(1)=stack3*(bex**dble(j0-lm(helixn)))*
     &                        scaleterm/
     &              ( stack3*(bex**dble(j0-lm(helixn)))*
     &                scaleterm+ed3ex(numseq(hm(helixn)),
     &            numseq(lm(helixn)),numseq(lm(helixn)+1))
     &            *u1(lm(helixn)+2, j0)*pscale(1)+
     &             u1(lm(helixn)+1, j0)-u1(lm(helixn)+2, j0)*pscale(1) )
          probb(2)=1.d0-probb(1)
!
          call discrete(nb, probb, irbn(1))
!
!            (l+1, en-1) unpaired, l=lm(helixn), j0=en-1
!                more output of unpaired bases if no more helix
          if (irbn(1) .eq. 1) then
            if ( (lm(helixn)+1) .le. j0) then
              do i=lm(helixn)+1, j0
                single(i)=1
                ltype(i)=4
              end do
            end if
            go to 100
          end if
!
          if (irbn(1) .eq. 2) then
            i0=lm(helixn)+1
            j0=j0
            go to 5
          end if
!
        end if
!
100     continue
        return
      end
!
      function augu(i, j, n, numseq)
!
        logical augu
        integer n, numseq(n), i, j
!
        augu = .FALSE.
        if ( ((numseq(i) .eq. 1) .and. (numseq(j) .eq. 4)) .or.
     &     ((numseq(i) .eq. 4) .and. (numseq(j) .eq. 1)) .or.
     &     ((numseq(i) .eq. 3) .and. (numseq(j) .eq. 4)) .or.
     &     ((numseq(i) .eq. 4) .and. (numseq(j) .eq. 3)) ) then
          augu = .TRUE.
        end if
        return
      end
!
!   2010-07-28 (Adam): See notes for this date, partf.f change 1a
!     function to evaluate hairpin enery with closing i-j
!     Moved to sample.f because it is used in both sample.f and partf.f
      function hpeex(i, j, n, numseq, ehex, eh4ex, ehsex,
     &               slopeex, hguex, h3cex, haex, hbex, tphbex)
!
        integer n, numseq(n), i, j, k, inc
        double precision hpeex, ehex(30), eh4ex(4,4,4,4,4,4),
     &   ehsex(4,4,4,4), slopeex, hguex, h3cex, haex, hbex, tphbex
        logical augu
!
        inc=j-i-1
!
        if (inc .gt. 30) then
          hpeex=ehex(30)*( slopeex**dlog(dble(inc)/30.d0) )
        else if (inc .eq. 4) then
!         reevaluate for tetra loop only (not triloop)
          hpeex=eh4ex(numseq(i),   numseq(i+1), numseq(i+2),
     &                numseq(i+3), numseq(i+4), numseq(i+5))
        else
          hpeex=ehex(inc)
        end if
!
!       add A-U, G-U penalty for triloops only
        if (inc .eq. 3) then
          if ( augu(i,j,n,numseq) ) then
            hpeex=hpeex*tphbex
          end if
        end if
!
!       add terminal stack if unpaired nt # >=4, p. 23 DHM et al 1998
        if (inc .ge. 4) then
          hpeex=hpeex*
     &      ehsex(numseq(i), numseq(i+1), numseq(j-1), numseq(j))
        end if
!
!      special GU bonus, 5' G preceded by 2 GG
        if ( (i-2) .ge. 1 .and. (j+2) .le. n ) then
          if ( numseq(i) .eq. 3 .and. numseq(j) .eq. 4 .and.
     &      numseq(i-1) .eq. 3 .and. numseq(i-2) .eq. 3 ) then
              hpeex=hpeex*hguex
          end if
        end if
!
!       3-c penalty
        if (inc .eq. 3) then
          if (numseq(i+1) .eq. 2 .and.
     &     numseq(i+2) .eq. 2 .and.
     &     numseq(i+3) .eq. 2) then
              hpeex=hpeex*h3cex
          end if
        end if
!
!       poly-c penalty
        if (inc .gt. 3) then
          do k=i+1, j-1
            if (numseq(k) .ne. 2) then
              go to 40
            end if
          end do
          hpeex=hpeex*(haex**dble(inc))*hbex
        end if
!
40      continue
        return
      end
!
