!!
!! sirna.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!    This is the F77 subroutine to compute the the probability weighted-
!    binding energy (ddeltag():delta delta G) between a siRNA and mRNA target
!    It also compute, delta G( deltag()), assuming the target
!    acts as a free oligomer (see Freier 1992) and only local disruption
!    (see mathews. et al 1999 RNA). Using Turner RNA/RNA parameters.
!    wi= variable window width (length of siRNA) for general siRNA
!      patterns; it is set at 21 now, including dTdT overhangs
!    einit: hybrid initiation penalty
!    e(I,J): stacking energy for
!      5'IJ 3'
!       I^cJ^c,
!    where I, J=A, C, G, U (1,2,3,4) and I^c, J^c are
!      complementary ribonucleotides
!!Comments: this was copied from soligo, call es()from main, then compute
!!        e(I,J) is better!!!
!      deltaG: binging dG for completely unpaired target seq.
!      ddeltaG: probability weighted dG
!      dg5as: stability of 5'end of antisense siRNA strand for nnn19
!      dg5s:  stability of 5'end of sense siRNA strand for nnn19
!      diffdg: dgas5-dgs5
!      intdg: stability of all pentamer susequences for nnn19
!      intaver: averaged stability for base 9-14 of antisense strand, nnn19
!      gcp: GC % of the siRNA (Not count overhang, Dharmacon convention)
!      sirna: sense of SiRNA; asirna: antisense;
!      nsirna: number of all SiRNA of length w; including 5', 3'UTR
!      sc: score
!      prsum: sum of unpaired probabilities for target bases in pairing
!        for now only consider 19 bases, overhang do not
!             to be TT or complementary to dinucleotide leader;
!        for TT, Dharmacon uses modification, won't pair with A?
!      duplexe: total energy of siRNA duplex, inc. A-AU terminal & dangling
!
!    2004/12/20 (Clarence):
!      Updated parameter list to include seqcopy and seqcopyc. these
!      two should only be used for sequence output purposes as these
!      two might contain the undetermined base 'N'.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!    2006/10/13 (Clarence):
!      Function paramter list updated to include new variables
!      outdir and outdirlen for the location to write output
!      files. See main.f for more details.
!

      subroutine ssirna(n, wi, nsirna , seq, seqc,
     &    psbase, psseq, ww, ed3, seqcopy, seqcopyc, outdir, outdirlen)

        integer n, wi, ww, i,j, id(n), num(n),gind, acuind
        integer nsirna

        integer scgc(nsirna), scau15(nsirna), scau16(nsirna),
     &          scau17(nsirna), scau18(nsirna), scau19(nsirna),
     &          scivr(nsirna), sca19(nsirna), sca3(nsirna),
     &          scu10(nsirna), scgc19(nsirna), scg13(nsirna),
     &          scdsse(nsirna), scais(nsirna), scfeature(nsirna),
     &          scthermo(nsirna), scaccess(nsirna), tscore(nsirna)

        double precision einit, e(4,4), ed3(4,4,4),
     &          psbase(n), psseq(ww,n), gcp, tphb,
     &          deltag(nsirna), ddeltag(nsirna),
     &          dg5as(nsirna), dg5s(nsirna), diffdg(nsirna),
     &          intdg(wi-2, nsirna), intaver(nsirna),
     &          prsum(nsirna), duplexe(nsirna)


        character*1 seq(n), seqc(n), seqcopy(n), seqcopyc(n),
     &           target(wi,nsirna),
     &          sirna(wi,nsirna),
     &          asirna(wi, nsirna)

!      set einit @ 4.1, see Intermolecular initiation free energy in
!            ~/Sfold/param/miscloop.dat
        data tphb /0.5d0/,
     &        einit /4.1d0/,
     &        ((e(i,j), j=1,4), i=1,4) /
     &         -0.9d0, -2.2d0, -2.1d0, -1.1d0,
     &         -2.1d0, -3.3d0, -2.4d0, -2.1d0,
     &         -2.4d0, -3.4d0, -3.3d0, -2.2d0,
     &         -1.3d0, -2.4d0, -2.1d0, -0.9d0  /

        character*(*) outdir
        integer outdirlen

!    2006/11/16 (Clarence): updated the way to pass filename to open().
!      both intel compiler and sun compiler take formats like
!      open(unit=2, file=(outdir(1:outdirlen) // '/sample.out'), but
!      gcc would fail for the filename part with the concatenation.
!      So we need to assign the filename to a variable first, then
!      pass it along to open(). bufstr would be the temp variable for
!      storing filename and buflen is its max length.
!
        integer buflen
        parameter(buflen=1000)
        character*(buflen) bufstr

!          output heading for filtered.out and stability.out
        bufstr = outdir(1:outdirlen) // '/sirna_f.out'
        open (unit=8, file=bufstr(1:len_trim(bufstr)), status='replace')
        bufstr = outdir(1:outdirlen) // '/stability.out'
        open (unit=9, file=bufstr(1:len_trim(bufstr)), status='replace')
        bufstr = outdir(1:outdirlen) // '/sirna_s.out'
        open (unit=1, file=bufstr(1:len_trim(bufstr)), status='replace')
        bufstr = outdir(1:outdirlen) // '/sirna.out'
        open (unit=2, file=bufstr(1:len_trim(bufstr)), status='replace')
        bufstr = outdir(1:outdirlen) // '/Dharmacon_thermo.out'
        open (unit=10, file=bufstr(1:len_trim(bufstr)), status='replace')

        do i=1, 15
          write (8, 11) '~'
        end do
        write (8, 85)
 85     format ('Output for siRNAs meeting all filter criteria', 1x, $)
        do i=1, 15
          write (8, 11) '~'
        end do
        write (8, 8)
        write (8, 86)
 86     format ( 'Line 1:' /
     &     'Column 1: target position (starting - ending)' /
     &     'Column 2: sense siRNA (5p --> 3p)' /
     &     'Column 3: antisense siRNA (5p --> 3p)' /
     &     'Column 4: dinucleotide leader preceding the target sequence' //
     &     'Line 2:' /
     &     'Column 1: total score for siRNA duplex' /
     &     'Column 2: target accessibility score' /
     &     'Column 3: duplex feature score' /
     &     'Column 4: duplex thermodynamics score' /
     &     'Column 5: siRNA GC content' /
     &     'Column 6: antisense siRNA binding energy (kcal/mol)' /
     &     'Column 7: differential stability of siRNA duplex ends
     &    (DSSE, in kcal/mol)' /
     &     'Column 8: average internal stability at the cleavage site
     &    (AIS, in kcal/mol)' /
     &     'Column 9: total stability of siRNA duplex (kcal/mol)' /
     &     'Column 10: sum of probabilities of unpaired target bases' /
     &     '           (column 4 of output file sstrand.out)'
     &          /  //
     &      'FILTER CRITERIA: ("<=": less than or equal to)' / /
     &      ' A) Antisense siRNA binding energy <= -10 kcal/mol
     &    (target accessibility rule);' /
     &      ' B) Duplex feature score of 6 or higher; ' /
     &      ' C) DSSE > 0 kcal/mol (asymmetry rule); ' /
     &      ' D) AIS > -8.6 kcal/mol (cleavage site instability rule); ' /
     &      ' E) 30% <= GC % <= 60%; ' /
     &      ' F) Exclusion of target sequence with at least one of AAAA,'/
     &      '    CCCC, GGGG, or UUUU.' / /
     &      'NOTES:' //
     &      ' 1) The starting (ending) position of the target sequence '/
     &      '    corresponds to position 19 (1) of the antisense siRNA ' /
     &      '    (i.e., dinucleotide leader and nt 22 and nt 23 in ' /
     &      '    Tuschl patterns are not considered by us to be part ' /
     &      '    of the target sequence); ' /
     &      ' 2) Sense siRNA=target sequence +3 prime dTdT overhang;'/
     &      '    dTdT for both sense and antisense siRNAs can be' /
     &      '    replaced by UU;'/
     &      ' 3) GC %= GC count in siRNA (excluding overhangs)/19 x100%;'
     &       /
     &      ' 4) DSSE = stability of 5p-antisense end of 4 base pairs ' /
     &      '          -stability of 5p-sense end of 4 base pairs; ' /
     &      '    the asymmetry rule is enforced by DSSE > 0.' /
     &      '    (see Schwartz et al. Cell, Vol. 115, 199-208, 2003). ' /
     &      ' 5) AIS = average of internal stability values for  ' /
     &      '    positions 9-14 of the antisense strand; starting at a ' /
     &      '    position, the internal stability is for 4 BP stacks;' /
     &      '    the rule of relative instability at the cleavage site' /
     &      '    is enforced by AIS > -8.6 kcal/mol, the midpoint ' /
     &      '    between the minimum of -3.6 and the maximum of -13.6. ' /
     &      '    (see Khvorova et al. Cell, Vol. 115, 209-216, 2003). ' /
     &      ' 6) Total siRNA duplex score is the sum of target '  /
     &      '    accessibility score, duplex feature score and' /
     &      '    duplex thermodynamics score, with a maximum of
     &    20 points;' /
     &      '    the accessibility score is based on antisense' /
     &      '    siRNA binding energy and has a range of [0, 8];' /
     &      '    the duplex feature score is computed with the' /
     &      '    algorithm by Reynolds et al. (Nature Biotech.,' /
     &      '    Vol. 22, 326-330, 2004), and has a range of [-2, 10];' /
     &      '    the duplex thermodynamics score has a range of  ' /
     &      '    [0, 2], with contribution of 1 point for DSSE > 0, ' /
     &      '    and 1 point for AIS > -8.6 kcal/mol.' /
     &            )


        do i=1, 80
          write (8, 11) '-'
        end do
        write (8, 8)
!                output header for score12.out
        do i=1, 15
          write (1, 11) '~'
        end do
        write (1, 15)
 15     format ('Output for siRNAs with total score of 12 or higher',1x,$)
        do i=1, 15
          write (1, 11) '~'
        end do
        write (1, 8)
        write (1, 16)
 16     format ( 'Line 1:' /
     &     'Column 1: target position (starting - ending)' /
     &     'Column 2: sense siRNA (5p --> 3p)' /
     &     'Column 3: antisense siRNA (5p --> 3p)' /
     &     'Column 4: dinucleotide leader preceding the target sequence' //
     &     'Line 2:' /
     &     'Column 1: total score for siRNA duplex' /
     &     'Column 2: target accessibility score' /
     &     'Column 3: duplex feature score' /
     &     'Column 4: duplex thermodynamics score' /
     &     'Column 5: siRNA GC content' /
     &     'Column 6: antisense siRNA binding energy (kcal/mol)' /
     &     'Column 7: differential stability of siRNA duplex ends
     &    (DSSE, in kcal/mol)' /
     &     'Column 8: average internal stability at the cleavage site
     &    (AIS, in kcal/mol)' /
     &     'Column 9: total stability of siRNA duplex (kcal/mol)' /
     &     'Column 10: sum of probabilities of unpaired target bases' /
     &     '           (column 4 of output file sstrand.out)'
     &          /  //
     &      'FILTER CRITERIA: ("<=": less than or equal to)' / /
     &      ' A) Total siRNA score of 12 points or higher; ' /
     &      ' B) Exclusion of target sequence with at least one of AAAA,'/
     &      '    CCCC, GGGG, or UUUU.' / /
     &      'NOTES:' //
     &      ' 1) The starting (ending) position of the target sequence '/
     &      '    corresponds to position 19 (1) of the antisense siRNA ' /
     &      '    (i.e., dinucleotide leader and nt 22 and nt 23 in ' /
     &      '    Tuschl patterns are not considered by us to be part ' /
     &      '    of the target sequence); ' /
     &      ' 2) Sense siRNA=target sequence +3 prime dTdT overhang;'/
     &      '    dTdT for both sense and antisense siRNAs can be' /
     &      '    replaced by UU;'/
     &      ' 3) GC %= GC count in siRNA (excluding overhangs)/19 x100%;'
     &       /
     &      ' 4) DSSE = stability of 5p-antisense end of 4 base pairs ' /
     &      '          -stability of 5p-sense end of 4 base pairs; ' /
     &      '    the asymmetry rule is enforced by DSSE > 0.' /
     &      '    (see Schwartz et al. Cell, Vol. 115, 199-208, 2003). ' /
     &      ' 5) AIS = average of internal stability values for  ' /
     &      '    positions 9-14 of the antisense strand; starting at a ' /
     &      '    position, the internal stability is for 4 BP stacks;' /
     &      '    the rule of relative instability at the cleavage site' /
     &      '    is enforced by AIS > -8.6 kcal/mol, the midpoint ' /
     &      '    between the minimum of -3.6 and the maximum of -13.6. ' /
     &      '    (see Khvorova et al. Cell, Vol. 115, 209-216, 2003). ' /
     &      ' 6) Total siRNA duplex score is the sum of target '  /
     &      '    accessibility score, duplex feature score and' /
     &      '    duplex thermodynamics score, with a maximum of
     &    20 points;' /
     &      '    the accessibility score is based on antisense' /
     &      '    siRNA binding energy and has a range of [0, 8];' /
     &      '    the duplex feature score is computed with the' /
     &      '    algorithm by Reynolds et al. (Nature Biotech.,' /
     &      '    Vol. 22, 326-330, 2004), and has a range of [-2, 10];' /
     &      '    the duplex thermodynamics score has a range of  ' /
     &      '    [0, 2], with contribution of 1 point for DSSE > 0, ' /
     &      '    and 1 point for AIS > -8.6 kcal/mol.' /
     &            )

        do i=1, 80
          write (1, 11) '-'
        end do
        write (1, 8)
!            output for all sirnas sirna.out
        do i=1, 25
          write (2, 11) '~'
        end do
        write (2, 25)
 25     format ('Output for all siRNAs', 1x, $)
        do i=1, 25
          write (2, 11) '~'
        end do
        write (2, 8)
        write (2, 26)
 26     format ( 'Line 1:' /
     &     'Column 1: target position (starting - ending)' /
     &     'Column 2: sense siRNA (5p --> 3p)' /
     &     'Column 3: antisense siRNA (5p --> 3p)' /
     &     'Column 4: dinucleotide leader preceding the target sequence' //
     &     'Line 2:' /
     &     'Column 1: total score for siRNA duplex' /
     &     'Column 2: target accessibility score' /
     &     'Column 3: duplex feature score' /
     &     'Column 4: duplex thermodynamics score' /
     &     'Column 5: siRNA GC content' /
     &     'Column 6: antisense siRNA binding energy (kcal/mol)' /
     &     'Column 7: differential stability of siRNA duplex ends
     &    (DSSE, in kcal/mol)' /
     &     'Column 8: average internal stability at the cleavage site
     &    (AIS, in kcal/mol)' /
     &     'Column 9: total stability of siRNA duplex (kcal/mol)' /
     &     'Column 10: sum of probabilities of unpaired target bases' /
     &     '           (column 4 of output file sstrand.out)'
     &          /  //
     &      'NOTES:' //
     &      ' 1) The starting (ending) position of the target sequence '/
     &      '    corresponds to position 19 (1) of the antisense siRNA ' /
     &      '    (i.e., dinucleotide leader and nt 22 and nt 23 in ' /
     &      '    Tuschl patterns are not considered by us to be part ' /
     &      '    of the target sequence); ' /
     &      ' 2) Sense siRNA=target sequence +3 prime dTdT overhang;'/
     &      '    dTdT for both sense and antisense siRNAs can be' /
     &      '    replaced by UU;'/
     &      ' 3) GC %= GC count in siRNA (excluding overhangs)/19 x100%;'
     &       /
     &      ' 4) DSSE = stability of 5p-antisense end of 4 base pairs ' /
     &      '          -stability of 5p-sense end of 4 base pairs; ' /
     &      '    the asymmetry rule is enforced by DSSE > 0.' /
     &      '    (see Schwartz et al. Cell, Vol. 115, 199-208, 2003). ' /
     &      ' 5) AIS = average of internal stability values for  ' /
     &      '    positions 9-14 of the antisense strand; starting at a ' /
     &      '    position, the internal stability is for 4 BP stacks;' /
     &      '    the rule of relative instability at the cleavage site' /
     &      '    is enforced by AIS > -8.6 kcal/mol, the midpoint ' /
     &      '    between the minimum of -3.6 and the maximum of -13.6. ' /
     &      '    (see Khvorova et al. Cell, Vol. 115, 209-216, 2003). ' /
     &      ' 6) Total siRNA duplex score is the sum of target '  /
     &      '    accessibility score, duplex feature score and' /
     &      '    duplex thermodynamics score, with a maximum of
     &    20 points;' /
     &      '    the accessibility score is based on antisense' /
     &      '    siRNA binding energy and has a range of [0, 8];' /
     &      '    the duplex feature score is computed with the' /
     &      '    algorithm by Reynolds et al. (Nature Biotech.,' /
     &      '    Vol. 22, 326-330, 2004), and has a range of [-2, 10];' /
     &      '    the duplex thermodynamics score has a range of  ' /
     &      '    [0, 2], with contribution of 1 point for DSSE > 0, ' /
     &      '    and 1 point for AIS > -8.6 kcal/mol.' /
     &            )

        do i=1, 80
          write (2, 11) '-'
        end do
        write (2, 8)
!                output stability data
        do i=1, 15
          write (9, 11) '~'
        end do
        write (9, 95)
 95     format ('siRNA ends and internal stability',
     &          1x, $)
        do i=1, 15
          write (9, 11) '~'
        end do
        write (9, 8)
        write (9, 96)
 96     format (
     &     'Line 1:  target position;',3x,'antisense siRNA (5p -> 3p);'
     &       / 9x,
     &      '5p-antisense stability(AntiS, in kcal/mol);', / 9x,
     &      '5p-sense stability (SS, in kcal/mol);',  / 9x,
     &      'differential stability of siRNA duplex ends
     &    (DSSE, in kcal/mol);', / 9x,
     &      'average internal stability at the cleavage site
     &    (AIS, in kcal/mol)' //
     &      'Line 2:  internal stability for antisense positions 1- 10)' /
     &      'Line 3:  internal stability for antisense positions 11- 19)'
     &        / /
     &      'NOTES:' //
     &      ' 1) AntiS is computed by a sum of energies for 4 base' /
     &      '    pair stacks and the 3-prime dangling T for the 5-prime' /
     &      '    end of the antisense siRNA strand; ' /
     &      '    SS is the sum for the 5-prime end of the sense strand;'/
     &      ' 2) DSSE = AntiS - SS; ' /
     &      '    the symmetry rule is enforced by DSSE > 0' /
     &      '    (see Schwartz et al. Cell, Vol. 115, 199-208, 2003); ' /
     &      ' 3) AIS = average of internal stability values for  ' /
     &      '    positions 9-14 of the antisense strand; starting at a ' /
     &      '    position, the internal stability is for 4 BP stacks;' /
     &      '    the rule of relative instability at the cleavage site' /
     &      '    is enforced by AIS > -8.6 kcal/mol, the midpoint ' /
     &      '    between the minimum of -3.6 and the maximum of -13.6 ' /
     &      '    (see Khvorova et al. Cell, Vol. 115, 209-216, 2003). ' /
     &           )

        do i=1, 80
          write (9, 11) '-'
        end do
        write (9, 8)
!      output heading for duplex features and thermodynamics
        do i=1, 10
          write (10, 11) '~'
        end do
        write (10, 105)
 105    format ('Individual scores for duplex features and thermodynamics'
     &          ,1x, $)
        do i=1, 10
          write (10, 11) '~'
        end do
        write (10, 8)
        write (10, 110)
 110    format(
     &     'Columns   1-2:  start position     end position ' /
     &     'Columns   3-7:  GC     A/U15   A/U16    A/U17   A/U18 ' /
     &     'Columns  8-12:  A/U19  inverse repeat   A19  A3  U10 ' /
     &     'Columns 13-15:  G/C19  G13     relative stability of
     &    siRNA duplex ends  ' /
     &     'Column     16:  siRNA duplex internal stability ' /
     &      )

        do i=1, 80
          write (10, 11) '-'
        end do
        write (10, 8)
!          convert mRNA base to numbering
        do i=1, n
          id(i)=i

          if(seqc(i) .eq. 'T' .or. seqc(i) .eq. 't') then
            seqc(i)='U'
          end if

!    2004/12/20 (Clarence):
!    added the following if case to change seqcopyc also
!
          if(seqcopyc(i) .eq. 'T' .or. seqcopyc(i) .eq. 't') then
            seqcopyc(i)='U'
          end if
          if(seq(i) .eq. 'A' .or. seq(i) .eq. 'a') then
            num(i)=1
            seq(i)='A'
          end if
          if(seq(i) .eq. 'C' .or. seq(i) .eq. 'c') then
            num(i)=2
            seq(i)='C'
          end if
          if(seq(i) .eq. 'G' .or. seq(i) .eq. 'g') then
            num(i)=3
            seq(i)='G'
          end if
          if(seq(i) .eq. 'U' .or. seq(i) .eq. 'u' .or.
     &         seq(i) .eq. 'T' .or. seq(i) .eq. 't') then
          num(i)=4
          seq(i)='U'
          end if
        end do

!      For general patterns (NN dinucleotife leader)

!      initialize counter for id of complete filtering output

        fid=0.0
        do 200 i=1, nsirna
          deltag(i)=einit
          ddeltag(i)=einit
          prsum(i)=0.d0

!      Below calculation only considers 19 target bases in
!      base bair with antisense siRNA with certainty; overhangs
!      not conisdered based on these difficulties 1) Dharmacon uses
!      modified TT that don't pair with A?? 2) one can use any overhang
!      either complementary or not to the dinucleotide leader
!      3) data in literature for non TT overhangs (Japan paper)
!      This takes effect 03/17/2004, giving slightl different results
!      for AA, NA leaders

!      use dinucleotide prob. to weigh 18 stacks

          do j=3, wi-1
            deltag(i)=deltag(i)+e(num(i+j-1), num(i+j))
            ddeltag(i)=ddeltag(i)+e(num(i+j-1), num(i+j))*psseq(2,i+j-1)
          end do

!      add A-U terminal penalty for sense base 1 (i+2) & base 19 (i+20)
!          weighted prob. of unpaired base 1 or 19
!          we do not consider dangling base, which depends on
!          target base @ i+1, and overhang used by user!

          if(seq(i+2) .eq. 'A' .or. seq(i+2) .eq. 'U' ) then
            deltag(i)=deltag(i)+tphb
            ddeltag(i)=ddeltag(i)+tphb*psbase(i+2)
          end if

          if(seq(i+20) .eq. 'A' .or. seq(i+20) .eq. 'U' ) then
            deltag(i)=deltag(i)+tphb
            ddeltag(i)=ddeltag(i)+tphb*psbase(i+20)
          end if

!      compute sum of unpaired probabilities of 19 targeted bases
          do j=3, wi
            prsum(i)=prsum(i)+psbase(i+j-1)
          end do

!!!!      below is currently unused code for TT overhang and assume pair with A
!      for AA leader, assume dTdT of antisense siRNA bind to AA
!      for BA leader (B=C,G,T), assume 3'-most dT free dangling
!      for NB leader,both dTdT free dangling(not relevant for
!      Tuschl patterns, N=A, C,G,T)

!      if(seq(i) .eq. 'A' .and. seq(i+1) .eq. 'A' ) then
!      do j=1, wi-1
!        deltag(i)=deltag(i)+e(num(i+j-1), num(i+j))
!        ddeltag(i)=ddeltag(i)+e(num(i+j-1), num(i+j))*psbase(i+j)
!      end do
!      end if

!      if(seq(i) .ne. 'A' .and. seq(i+1) .eq. 'A') then
!      do j=2, wi-1
!        deltag(i)=deltag(i)+e(num(i+j-1), num(i+j))
!        ddeltag(i)=ddeltag(i)+e(num(i+j-1), num(i+j))*psbase(i+j)
!      end do
!      end if

!      if(seq(i+1) .ne. 'A') then
!      do j=3, wi-1
!        deltag(i)=deltag(i)+e(num(i+j-1), num(i+j))
!        ddeltag(i)=ddeltag(i)+e(num(i+j-1), num(i+j))*psbase(i+j)
!        prsum(i)=prsum(i)+psbase(i+j)
!      end do
!      end if


!      compute stability of 5' end of both sense and antisense strand
!      by summing of 4 stacks and 3' dangling dT (treated as U)
!            siRNA base position=i+2 (for dinucleotide leader)
!      for W-C pair sum of num=5;

          dg5s(i)=e(num(i+2), num(i+3))+e(num(i+3), num(i+4))+
     &              e(num(i+4), num(i+5))+e(num(i+5), num(i+6))+
     &              ed3(num(i+2), 5-num(i+2), 4)

!      add terminal A-U penalty (not used in Zamore calculation!)
          if(seq(i+2) .eq. 'A' .or. seq(i+2) .eq. 'U' ) then
            dg5s(i)=dg5s(i)+tphb
          end if

          dg5as(i)=e(num(i+16), num(i+17))+e(num(i+17), num(i+18))+
     &               e(num(i+18), num(i+19))+e(num(i+19), num(i+20))+
     &               ed3(5-num(i+20),num(i+20), 4)

!      add terminal A-U penalty (not used in Zamore calculation!)
          if(seq(i+20) .eq. 'A' .or. seq(i+20) .eq. 'U' ) then
            dg5as(i)=dg5as(i)+tphb
          end if

          diffdg(i)=dg5as(i)-dg5s(i)

!          compute the total free energy of siRNA duplex
!          including A-U, 3' dnagling
          duplexe(i)=dg5s(i)+dg5as(i)
          do j=6, 15
            duplexe(i)=duplexe(i)+e(num(i+j), num(i+j+1))
          end do

!          compute internal stability (sum of 4 BP stack) of pentamer
!      subsequences (done by Khvorova et al) for each of 1-19 postion
!      with window shifting from 5' to 3' of the antisense strand
!      except for a) for last 4 bases of 3' end, we reverse the
!      shifting direction so the results are symmetric for the ends;
!      as a result intdg(j,i)=intdg(j-4,i), j=16,17, 18, 19
!      they "extended" AS siRNA with target mRNA sequence, not good!;
!      b) for postion 1 & 19, use dangling and A-U terminal penalty

          intdg(1,i)=dg5as(i)

          do j=2, wi-6
            intdg(j,i)=e(num(i+16-(j-1)), num(i+17-(j-1)))+
     &                    e(num(i+17-(j-1)), num(i+18-(j-1)))+
     &                    e(num(i+18-(j-1)), num(i+19-(j-1)))+
     &                    e(num(i+19-(j-1)), num(i+20-(j-1)))
          end do

!      below is same as:     intdg(j,i)=intdg(j-4,i)

          intdg(wi-2,i)=dg5s(i)

          do j=wi-5, wi-3
!           intdg(j,i)=intdg(j-4,i)
            intdg(j,i)=e(num(i+2+((wi-2)-j)), num(i+3+((wi-2)-j)))+
     &                    e(num(i+3+((wi-2)-j)), num(i+4+((wi-2)-j)))+
     &                    e(num(i+4+((wi-2)-j)), num(i+5+((wi-2)-j)))+
     &                    e(num(i+5+((wi-2)-j)), num(i+6+((wi-2)-j)))
          end do

!      compute the average of internal stability (6 terms) for
!      base 9-14 of antisense strand
!      Ye's note: I would rather computer the dg for 5 stack of
!      the helix segmnet, I did not so it is consitent with their work

          intaver(i)=0.d0
          do j =9, 14
            intaver(i)=intaver(i)+intdg(j,i)/6.d0
          end do

!            Antisense siRNA & target motif: 5' -> 3'
!            was asirna(j,i)=seqc(i+wi-1-(j-1)) for AA(N19)

!    2004/12/20 (Clarence):
!    Updated the following codes to use seqcopy and
!    seqcopyc to set asirna and target
!
          do j=1, wi-2
            asirna(j,i)=seqcopyc(i+wi-1-(j-1))
          end do
          asirna(wi-1,i)='T'
          asirna(wi, i)='T'

          do j=1, wi
            target(j,i)=seqcopy(i+j-1)
          end do
!          GC % of siRNA, excluding overhangs, to be consistent
!      with  Dharmacon usage in their Nature Biotch paper

          gcp=0.d0
          do j=3, wi
            if(target(j,i) .eq. 'G' .or. target(j,i) .eq. 'C') then
              gcp=gcp+1.d0/dble(wi-2)*100.d0
            end if
          end do

!    2004/12/20 (Clarence):
!    Updated the following codes to use seqcopy
!    to set sirna
!

!          sense siRNA:5' -> 3'     (target(j+2, i) for AA(N19))
          do j=1, wi-2
            sirna(j,i)=seqcopy(i+j-1+2)
          end do
          sirna(wi-1,i)='T'
          sirna(wi, i)='T'

!!          compute indicator for GGGG , and AAAA/CCCC/UUUU
!!          note: for reporting the case of  at least one
!          homopolymer, the following is can be redundant,
!          say GGGGG get assigned 1 twice, but this is ok.
          gind=0
          acuind=0

          do j=1, wi-3
            if(seq(i+j-1) .eq. 'G' .and.
     &            seq(i+j) .eq. 'G' .and.
     &            seq(i+j+1) .eq. 'G' .and.
     &            seq(i+j+2) .eq. 'G') then
              gind=1
            end if

            if(seq(i+j-1) .eq. 'A' .and.
     &            seq(i+j) .eq. 'A' .and.
     &            seq(i+j+1) .eq. 'A' .and.
     &            seq(i+j+2) .eq. 'A') then
              acuind=1
            end if

            if(seq(i+j-1) .eq. 'C' .and.
     &            seq(i+j) .eq. 'C' .and.
     &            seq(i+j+1) .eq. 'C' .and.
     &            seq(i+j+2) .eq. 'C') then
              acuind=1
            end if

            if(seq(i+j-1) .eq. 'U' .and.
     &            seq(i+j) .eq. 'U' .and.
     &            seq(i+j+1) .eq. 'U' .and.
     &            seq(i+j+2) .eq. 'U') then
              acuind=1
            end if
          end do

!!          now compute all the scores
!!      Part 1:  feature scores by Reynolds et al 2004 Nat. Biot.
!           using sense strand sirna(j, i)
!                30 <= GC% <= 52
          if((gcp .ge. 30) .and. (gcp .le. 52)) then
            scgc(i)=1
          else
            scgc(i)=0
          end if
!                A or U at 15
          if(sirna(15, i) .eq. 'A' .or. sirna(15, i) .eq. 'U') then
            scau15(i)=1
          else
            scau15(i)=0
          end if
!                A or U at 16
          if(sirna(16, i) .eq. 'A' .or. sirna(16, i) .eq. 'U') then
            scau16(i)=1
          else
            scau16(i)=0
          end if
!                A or U at 17
          if(sirna(17, i) .eq. 'A' .or. sirna(17, i) .eq. 'U') then
            scau17(i)=1
          else
            scau17(i)=0
          end if
!                A or U at 18
          if(sirna(18, i) .eq. 'A' .or. sirna(18, i) .eq. 'U') then
            scau18(i)=1
          else
            scau18(i)=0
          end if
!                A or U at 19
          if(sirna(19, i) .eq. 'A' .or. sirna(19, i) .eq. 'U') then
            scau19(i)=1
          else
            scau19(i)=0
          end if
!                A19
          if(sirna(19, i) .eq. 'A') then
           sca19(i)=1
          else
            sca19(i)=0
          end if
!                A3
          if(sirna(3, i) .eq. 'A') then
            sca3(i)=1
          else
            sca3(i)=0
          end if
!                U10
          if(sirna(10, i) .eq. 'U') then
            scu10(i)=1
          else
            scu10(i)=0
          end if
!                G19 or C19 (-1 penalty)

          if(sirna(19, i) .eq. 'G' .or. sirna(19, i) .eq. 'C') then
            scgc19(i)=-1
          else
            scgc19(i)=0
          end if
!                G13 (-1 penalty)
          if(sirna(13, i) .eq. 'G') then
            scg13(i)=-1
          else
            scg13(i)=0
          end if

!         inverted repeat of 4 nt,separated by >= 4 bases (stable Hairpin)
!          for now, Tm calculation is missing!

          scivr(i)=1
          do j=1, 8
            do k=j+8, 16
              if(sirna(j, i) .eq. asirna(19-(k+3)+1,i) .and.
     &               sirna(j+1, i) .eq. asirna(19-(k+2)+1,i) .and.
     &               sirna(j+2, i) .eq. asirna(19-(k+1)+1,i) .and.
     &               sirna(j+3, i) .eq. asirna(19-k+1,i) ) then
                scivr(i)=0
!            print *, i+2, i+wi-1, j, k, i+2+j-1, i+2+k-1
                go to 101
              end if
            end do
          end do
 101      continue
!                DSSE

          if(diffdg(i) .gt. 0.d0) then
            scdsse(i)=1
          else
           scdsse(i)=0
          end if
!                AIS
          if(intaver(i) .gt. (-8.6d0) )then
            scais(i)=1
          else
            scais(i)=0
          end if
!                accessibility

!    for comple helix w/ prob. 1, siRNA binding is 4.1= initiation

          if(ddeltag(i) .gt. 0.d0 )then
            scaccess(i)=0
          end if

          if(ddeltag(i) .le. 0.d0 .and. ddeltag(i) .gt. (-2.d0))then
            scaccess(i)=0
          end if

          if(ddeltag(i) .le. (-2.d0) .and. ddeltag(i) .gt. (-4.d0))then
            scaccess(i)=1
          end if

          if(ddeltag(i) .le. (-4.d0) .and. ddeltag(i) .gt. (-6.d0))then
            scaccess(i)=2
          end if

          if(ddeltag(i) .le. (-6.d0) .and. ddeltag(i) .gt. (-8.d0))then
            scaccess(i)=3
          end if

          if(ddeltag(i) .le. (-8.d0) .and. ddeltag(i) .gt. (-10.d0))then
            scaccess(i)=4
          end if

          if(ddeltag(i) .le. (-10.d0) .and. ddeltag(i) .gt. (-12.d0))then
            scaccess(i)=5
          end if

          if(ddeltag(i) .le. (-12.d0) .and. ddeltag(i) .gt. (-14.d0))then
            scaccess(i)=6
          end if

          if(ddeltag(i) .le. (-14.d0) .and. ddeltag(i) .gt. (-16.d0))then
            scaccess(i)=7
          end if

          if(ddeltag(i) .le. (-16.d0)) then
            scaccess(i)=8
          end if

!          score sums and total score
          scfeature(i)=scgc(i)+scau15(i)+scau16(i)+
     &                   scau17(i)+scau18(i)+scau19(i)+
     &                   scivr(i)+sca19(i)+sca3(i)+
     &                   scu10(i)+scgc19(i)+scg13(i)

          scthermo(i)=scdsse(i)+scais(i)
          tscore(i)=scfeature(i)+scthermo(i)+scaccess(i)

!!********outputs
!          filtered.out for using all key filters
!          regardless of dinucleotide leader
!          for average internal stability of AS strand nt9-14
!           -8.6 is the midpoint of min=-3.6, max=-13.6
!             Dhramacon score >=6
!            complete major filters defined by

          if((gcp .ge. 30.d0) .and.
     &         (gcp .le. 60.d0) .and.
     &         (ddeltag(i) .le. -10.d0) .and.
     &         (gind .eq. 0) .and.
     &         (acuind .eq. 0) .and.
     &         (diffdg(i) .gt. 0.d0) .and.
     &         (intaver(i) .gt. -8.6d0) .and.
     &         (scfeature(i) .ge. 6) ) then

            write (8, 20) i+2, i+wi-1, ' '
            write (8, 11) ' '

            do j=1, wi
              write (8, 11) sirna(j, i)
            end do

            write (8, 11) ' '
            write (8, 11) ' '

            do j=1, wi
              write (8, 11) asirna(j, i)
            end do

            write (8, 11) ' '
            write (8, 11) ' '

            write (8, 11) seqcopy(i)
            write (8, 21) seqcopy(i+1)
            write (8, 11) ' '

            write (8, 70) tscore(i), scaccess(i), scfeature(i),
     &                       scthermo(i), gcp, ddeltag(i),
     &                       diffdg(i), intaver(i), duplexe(i),prsum(i)
          end if

!      output  those with total score >=12
!      and no 4 G,C A< U to file score12.out

          if(tscore(i) .ge. 12 .and.
     &         (gind .eq. 0) .and.
     &         (acuind .eq. 0) ) then

            write (1, 20) i+2, i+wi-1, ' '
            write (1, 11) ' '

            do j=1, wi
              write (1, 11) sirna(j, i)
            end do

            write (1, 11) ' '
            write (1, 11) ' '

            do j=1, wi
              write (1, 11) asirna(j, i)
            end do

            write (1, 11) ' '
            write (1, 11) ' '

            write (1, 11) seqcopy(i)
            write (1, 21) seqcopy(i+1)
            write (1, 11) ' '

            write (1, 70) tscore(i), scaccess(i), scfeature(i),
     &                       scthermo(i), gcp, ddeltag(i),
     &                       diffdg(i), intaver(i), duplexe(i),prsum(i)

          end if

!                output all siRNAs

!                individual duplex feature & thermo scores
          write (10, 120) i+2, i+wi-1,
     &                      scgc(i), scau15(i), scau16(i),
     &                      scau17(i),scau18(i), scau19(i),
     &                      scivr(i), sca19(i), sca3(i),
     &                      scu10(i), scgc19(i), scg13(i),
     &                      scdsse(i), scais(i)

          write (2, 20) i+2, i+wi-1, ' '
          write (2, 11) ' '

          do j=1, wi
            write (2, 11) sirna(j, i)
          end do

          write (2, 11) ' '

          do j=1, wi
            write (2, 11) asirna(j, i)
          end do

          write (2, 11) ' '
          write (2, 11) ' '

          write (2, 11) seqcopy(i)
          write (2, 21) seqcopy(i+1)
          write (2, 11) ' '

          write (2, 70) tscore(i), scaccess(i), scfeature(i),
     &                    scthermo(i), gcp, ddeltag(i),
     &                    diffdg(i), intaver(i), duplexe(i), prsum(i)

!            stability. out for all possible targets
!      output target position, AS strand, 5' & 3' (w/ dangling)
!          difference, average 9-14
!      line 2: internal stabilities for 19 AS strand positions
          write (9, 20) i+2, i+wi-1, ' '

          do j=1, wi
            write (9, 11) asirna(j, i)
          end do

          write (9, 63)  dg5as(i), dg5s(i), diffdg(i), intaver(i)

          write(9, 64) (intdg(j, i), j=1, 10)
          write(9, 65) (intdg(j, i), j=11, wi-2)

 200      continue

 8        format (/ )
 11       format(A1, $)
 21       format(A1)
 20       format(i5, '-', i5, A1, $)
 63       format(4(1x,f5.1))
 64       format(12x, f5.1, 9(1x, f5.1))
 65       format(12x, f5.1, 8(1x, f5.1))
 70       format(12x, I2, 2x, I1, 2x, I2, 2x, I1,2x, f4.1,'%',
     &       2x, f5.1, 3(2x, f5.1), 1x, f6.2)

 120      format(i5, '-', i5, 2x, 10(i1, 1x), 2(i2, 1x), 2(i1, 1x))

          close(1)
          close(2)
          close(8)
          close(9)

          return
        end
