!!
!! soligo.f
!!
!! principal developer: Dr. Ye Ding
!! modifications to this source code are only permitted with the permission of the copyright holder
!! or the principal developer.
!!
!! Copyright Health Resources Inc. 2007 
!! You may use distribute, and modify, this code under the terms
!! of the SFold license.  A copy of the Sfold license should have
!! been distribute,d with this file.
!!
!!
!  This is the F77 subroutine to compute the the probability weighted-
!  binding energy (ddeltag():delta delta G) between an oligo and mRNA target
!  It also compute, delta G( deltag()), assuming the target
!  acts as a free oligomer (see Freier 1992) and only local disruption
!  (see mathews. et al 1999 RNA). Using DNA/RNA parameters.
!  We now only output ddeltag()
!  w= window width (length of oligo)
!  einit: hybrid initiation penalty
!  e(I,J): stacking energy for
!    rIJ
!    dI^cJ^c,
!  where I, J=A, C, G, U (1,2,3,4) and I^c, J^c are DNA complements
!
!    2004/12/20 (Clarence):
!  Updated parameter list to include seqcopy and seqcopyc. these
!  two should only be used for sequence output purposes as these
!  two might contain the undetermined base 'N'.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!    2006/10/13 (Clarence):
!       Function paramter list updated to include new variables
!       outdir and outdirlen for the location to write output
!       files. See main.f for more details.
      subroutine soligo(n, w, noligo, seq, seqc,
     &       psbase, psseq, ww, seqcopy, seqcopyc, outdir, outdirlen)

        integer n, w, ww, id(n), num(n), noligo, i, j, gind

        double precision einit, e(4,4), psbase(n),
     &       gcp, deltag(noligo), ddeltag(noligo),
     &       psseq(ww,n)

        character*1 seq(n), seqc(n), seqcopy(n), seqcopyc(n)
        character*1 oligo(w, noligo), target(w, noligo)
!  character*20 oligo(noligo), target(noligo)

        data  einit / 3.1d0/,
     &        ((e(i,j), j=1,4), i=1,4) /
     &       -1.d0, -2.1d0, -1.8d0, -0.9d0,
     &       -0.9d0, -2.1d0, -1.7d0, -0.9d0,
     &       -1.3d0, -2.7d0, -2.9d0, -1.1d0,
     &        -0.6d0, -1.5d0, -1.6d0, -0.2d0  /

        character*(*) outdir
        integer outdirlen

!    2006/11/16 (Clarence): updated the way to pass filename to open().
!       both intel compiler and sun compiler take formats like
!       open(unit=2, file=(outdir(1:outdirlen) // '/sample.out'), but
!       gcc would fail for the filename part with the concatenation.
!       So we need to assign the filename to a variable first, then
!       pass it along to open(). bufstr would be the temp variable for
!       storing filename and buflen is its max length.
!
        integer buflen
        parameter(buflen=1000)
        character*(buflen) bufstr

!      ouput w/o and with filer

        bufstr = outdir(1:outdirlen) // '/oligo.out'
        open (unit=6, file=bufstr(1:len_trim(bufstr)),
     &          status='replace')
        bufstr = outdir(1:outdirlen) // '/oligo_f.out'
        open (unit=7, file=bufstr(1:len_trim(bufstr)),
     &          status='replace')

!       output heading,  no filter
        do i=1, 20
          write (6, 11) '~'
        end do

        write (6, 5)
5       format ('Output for design of antisense oligos', 1x, $)

        do i=1, 20
          write (6, 11) '~'
        end do

        write (6, 8)

        write (6, 9)
9       format(
     &     'Column 1: target position (starting - ending)' /
     &     'Column 2: target sequence (5p --> 3p)' /
     &     'Column 3: antisense oligo (5p --> 3p)' /
     &     'Column 4: GC content' /
     &     'Column 5: oligo binding energy (kcal/mol)' /
     &     'Column 6: GGGG indicator ' / /
     &     'NOTES:' //
     &     'GGGG indicator=1 for at least one GGGG in the target'/
     &     'sequence; indicator=0 otherwise.' )

        do i=1, 80
          write (6, 11) '-'
        end do

        write (6, 8)

!       output heading,  with filter
        do i=1, 17
          write (7, 11) '~'
        end do

        write (7, 75)

75      format ('Filtered output for design of antisense oligos', 1x, $)

        do i=1, 17
          write (7, 11) '~'
        end do

        write (7, 8)

        write (7, 79)
79      format(
     &     'Column 1: target position (starting - ending)' /
     &     'Column 2: target sequence (5p --> 3p)' /
     &     'Column 3: antisense oligo (5p --> 3p)' /
     &     'Column 4: GC content' /
     &     'Column 5: oligo binding energy (kcal/mol)'  //
     &      'FILTER CRITERIA: ("<=": less than or equal to)' / /
     &      ' A) 40% <= GC % <= 60%;' /
     &      ' B) Antisense oligo binding energy <= -8 kcal/mol;' /
     &      ' C) No GGGG in the target sequence.' )

        do i=1, 80
          write (7, 11) '-'
        end do

        write (7, 8)

8       format (/ )

!      convert mRNA base to numbering
        do i=1, n

          id(i)=i

          if (seq(i) .eq. 'A' .or. seq(i) .eq. 'a') then
            num(i)=1
            seq(i)='A'
          end if
          if (seq(i) .eq. 'C' .or. seq(i) .eq. 'c') then
            num(i)=2
            seq(i)='C'
          end if
          if (seq(i) .eq. 'G' .or. seq(i) .eq. 'g') then
            num(i)=3
            seq(i)='G'
          end if
          if (seq(i) .eq. 'U' .or. seq(i) .eq. 'u'
     & .or. seq(i) .eq. 'T' .or. seq(i) .eq. 't') then
          num(i)=4
          seq(i)='U'
          end if

        end do

!    compute  delta G and delta delta G
        do i=1, noligo
          deltag(i)=einit
          ddeltag(i)=einit

!    use dinucleotide prob. to weigh (w-1) stacks

          do j=1, w-1
            deltag(i)=deltag(i)+e(num(i+j-1), num(i+j))
            ddeltag(i)=ddeltag(i)+e(num(i+j-1), num(i+j))*psseq(2,i+j-1)
          end do

!    add A-U terminal penalty for ends, using RNA:RNA parameters?
!    RNA:DNA not available, weighted by the prob. of unpaired ends
!    below not used, for a) no confidence in using RNA:RNA parameters
!    b) for Turner's way of measuring, they want A-U, not clear
!    why this is also the case for Sugimoto et al' RNA:DNA

!     if (seq(i) .eq. 'A' .or. seq(i) .eq. 'U' ) then
!       deltag(i)=deltag(i)+tphb
!       ddeltag(i)=ddeltag(i)+tphb*psbase(i)
!     end if

!     if (seq(i+w-1) .eq. 'A' .or. seq(i+w-1) .eq. 'U' ) then
!       deltag(i)=deltag(i)+tphb
!       ddeltag(i)=ddeltag(i)+tphb*psbase(i+w-1)
!     end if

!        oligo: 5' -> 3'

!    2004/12/20 (Clarence):
!  updated the following loop to set strings in oligo
!  and target to copy from seqcopy and seqcopyc
!  instead of the modified sequence. It should not
!  affect other parts of the code where seq is still
!  being used
!
          do j=1, w
            oligo(j, i)=seqcopyc(i+w-1-(j-1))
            target(j, i)=seqcopy(i+j-1)
!        oligo(i)(j:j)=seqc(i+w-1-(j-1))
!        target(i)(j:j)=seq(i+j-1)
          end do

!        GC % of target seq.
          gcp=0.d0
          do j=1, w
            if (target(j,i) .eq. 'G' .or. target(j,i) .eq. 'C') then
              gcp=gcp+1.d0/dble(w)*100.d0
            end if
          end do

!!      compute indicator for GGGG ,
!!      note: for reporting the case of  at least one
!      homopolymer, the following is can be redundant,
!      say GGGGG get assigned 1 twice, but this is ok.
          gind=0

          do j=1, w-3

            if(seq(i+j-1) .eq. 'G' .and.
     &        seq(i+j) .eq. 'G' .and.
     &        seq(i+j+1) .eq. 'G' .and.
     &        seq(i+j+2) .eq. 'G') then
            gind=1
            end if

          end do
!        output w/o filer

          write(6, 10) i, i+w-1, ' '

          do j=1, w
            write(6, 11) target(j, i)
          end do

          write (6, 11) ' '

          do j=1, w
            write(6, 11) oligo(j, i)
          end do
          write(6, 13)  gcp, ddeltag(i), gind

!        output  with filter defined by if:
          if(   (gcp .ge. 40.d0) .and. (gcp .le. 60.d0) .and.
     &              (ddeltag(i) .le. -8.d0) .and.
     &        (gind .eq. 0)   ) then

            write(7, 10) i, i+w-1, ' '

            do j=1, w
              write(7, 11) target(j, i)
            end do

            write (7, 11) ' '

            do j=1, w
              write(7, 11) oligo(j, i)
            end do
            write(7, 14) gcp, ddeltag(i)

          end if
        end do

10      format(i4,'-', i4, A1, $)
11      format(A1, $)
13      format(2x,f4.1,'%', 2x, f5.1, 2x, i1)
14      format(2x,f4.1,'%', 2x, f5.1)

        close (6)
        close(7)

        return
      end
